import atexit
import os
import sys
import uuid
from types import FunctionType

from werkzeug.exceptions import NotFound as SuperNotFound
from werkzeug.routing import Map, Rule

from restfx.globals import _app_ctx_stack
from restfx.util import ContextStore, utils
from .http import HttpRequest, NotFound, ServerError
from .middleware import MiddlewareManager
from .routes import Collector


class AppContext:
    def __init__(self, app):
        self.app = app
        self.store = ContextStore(_app_ctx_stack)
        self.ref_count = 0

    def push(self):
        top = _app_ctx_stack.top
        if not top or top != self:
            _app_ctx_stack.push(self)
        self.ref_count += 1

    def pop(self):
        if _app_ctx_stack.top != self:
            return
        self.ref_count -= 1
        if self.ref_count > 0:
            return
        _app_ctx_stack.pop()
        del self.store

    def __enter__(self):
        self.push()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()


class App:
    # APP 实例集合
    _APPS = {}

    def __init__(self,
                 app_root: str,
                 app_id: str = None,
                 api_prefix='api',
                 debug=True,
                 append_slash=False,
                 strict_mode=False,
                 api_page_enabled=None,
                 api_page_name: str = None,
                 api_page_expanded=False,
                 api_page_cache=None,
                 api_page_addition=None,
                 api_page_header=None,
                 api_page_footer=None,
                 api_page_assets: tuple = None,
                 ):
        """

        :param app_root: 应用的根目录
        :param app_id: 全局的唯一 id, 用于标识一个APP。
        :param api_prefix: API接口URL前缀
        :param debug: 是否启用调试模式
        :param append_slash: 是否在URL末尾使用 / 符号
        :param strict_mode: 是否启用严格模式。在严格模式下，不允许在请求时携带未声明的参数
        :param api_page_enabled: 只否启用接口页面
        :param api_page_name: 接口页面名称
        :param api_page_expanded: 是否展示接口列表
        :param api_page_cache: 是否缓存接口数据
        :param api_page_addition: 接口页面上要展示的接口的附加信息函数，其接收一个 dict 类型的参数 route_info
        :param api_page_header: 接口页面上，需要自定义展示在页面顶部的信息
        :param api_page_footer: 接口页面上，需要自定义展示在页面底部的信息
        :param api_page_assets: 接口页面上，需要自定义的JS和CSS以及其它资源 (通过 static 托管，此接口仅填写地址)
        """
        from .config import AppConfig
        from .routes import ApiPage
        from .routes import Router
        from .util import Logger

        self.id = app_id or str(uuid.uuid4())
        self._logger = Logger(app_id)

        self._APPS[self.id] = self

        self.config = AppConfig(self.id, app_root, debug, append_slash,
                                strict_mode, api_page_enabled, api_page_name,
                                api_page_expanded, api_page_cache, api_page_addition,
                                api_page_header, api_page_footer, api_page_assets)
        self.config.middleware_manager = MiddlewareManager(self.id, self.config)
        self._api_prefix = api_prefix
        self._router = Router(self.config)
        self._api_page = ApiPage(self.config)

        self._custom_url_map = {}

        self._url_map = Map([
            Rule('/%s%s' % (api_prefix, '/' if append_slash else ''), endpoint='_api_page'),
            Rule('/%s/<path:entry>%s' % (api_prefix, '/' if append_slash else ''), endpoint='entry_only')
        ])

        Collector.create(app_id, app_root, append_slash)

        self.context = AppContext(self)
        self.context.push()

        super(App, self).__init__()

    def dispose(self):
        self.config.middleware_manager.handle_shutdown()

        del self.config.middleware_manager

        self._logger.info('App "%s" is shutting down' % self.id)
        Collector.destroy(self.id)
        self.context.pop()
        if self.id in self._APPS:
            del self._APPS[self.id]

    def handle_wsgi_request(self, environ, start_response):
        """
        接收并处理来自 wsgi 的请求
        :param environ:
        :param start_response:
        :return:
        """

        request = None
        response = None
        try:
            request = HttpRequest(environ, self)
            request.context().push()
            self.config.middleware_manager.handle_coming(request)

            adapter = self._url_map.bind_to_environ(environ)

            endpoint, values = adapter.match()
            if endpoint == '_api_page':
                response = self._api_page.dispatch(request)
            elif endpoint == 'entry_only':
                response = self._router.dispatch(request, values['entry'])
            elif endpoint in self._custom_url_map:
                response = self._custom_url_map[endpoint](request, **values)
            else:
                response = NotFound()
        except Exception as e:
            msg = utils.get_exception_info(e)
            if request:
                msg = 'Error occurred during handling request %r:\n\t%s' % (request.path, msg)

            self._logger.error(msg)

            if isinstance(e, SuperNotFound):
                response = NotFound()
            elif self.config.debug:
                response = ServerError(msg.replace('<', '&lt;').replace('>', '&gt;'))
            else:
                response = ServerError()
        finally:
            self.config.middleware_manager.handle_leaving(request, response)
            request.context().pop()
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        """
        wsgi 入口
        :param environ:
        :param start_response:
        :return:
        """
        if not self.config.static_map:
            return self.handle_wsgi_request(environ, start_response)

        from werkzeug.middleware.shared_data import SharedDataMiddleware
        return SharedDataMiddleware(self.handle_wsgi_request, self.config.static_map)(
            environ, start_response)

    def startup(self, host=None, port=9127, threaded=True, **kwargs):
        """
        开发时使用的服务器，一般来说，应仅用于开发
        :param host:
        :param port:
        :param threaded:
        :param kwargs: 适用于 werkzeug 的 run_simple 函数的其它参数
        :return:
        """
        from restfx.util import helper
        from werkzeug.serving import run_simple

        debug = self.config.debug

        if debug:
            helper.print_meta('dev-server *DEBUG MODE*')
        else:
            helper.print_meta('dev-server')

        # 检查启动参数
        argv = sys.argv
        if len(argv) > 1:
            server_arg = argv[1].split(':')
            if len(server_arg) == 1:
                self._logger.warning('Invalid startup argument "%s" (ignored):' % argv[1])
            else:
                host = server_arg[0]
                port = int(server_arg[1])

        # 检查环境变量
        env_host = os.environ.get('RESTFX_HOST')
        env_port = os.environ.get('RESTFX_PORT')

        if env_host is not None:
            host = env_host
        if env_port is not None:
            port = int(env_port)

        # 支持 空串和 * 标志
        if host in [None, '', '*']:
            host = '0.0.0.0'

        if self.config.api_page_enabled:
            if host == '0.0.0.0':
                from .util import utils
                ips = utils.get_ip_list()
            else:
                ips = [host]

            print(' * Table of APIs:')
            for ip in ips:
                print('\t- http://%s:%s/%s%s' % (
                    ip, port, self._api_prefix, '/' if self.config.append_slash else ''
                ))

        run_simple(host, port, self, use_debugger=debug, use_reloader=debug, threaded=threaded, **kwargs)

    def register_types(self, *types):
        """
        注册路由参数上使用的类型，目前仅支持枚举类型
        :param types:
        :return:
        """
        collector = Collector.get(self.id)
        for type_item in types:
            collector.global_types[type_item.__name__] = type_item

    def collect(self):
        """
        收集路由
        :return:
        """
        return Collector.get(self.id).collect(self.config.routes_map)

    def persist(self, filename: str = '', encoding='utf8'):
        """
        持久化路由信息到文件
        :param filename: 不指定此值时，此函数会返回路由文件内容
        :param encoding:
        :return:
        """
        return Collector.get(self.id).persist(self.config.routes_map, filename, encoding)

    def set_logger(self, logger):
        """
        设置自定义的 logger
        :param logger:
        :return:
        """
        self._logger.custom_logger = logger
        return self

    def map_urls(self, urls_map: dict):
        """
        自定义的 url 映射，这些映射不会经过路由处理，而是直接来自 werkzeug
        :param urls_map:
        :return:
        """
        for url in urls_map:
            self._custom_url_map[url] = urls_map[url]
            self._url_map.add(Rule(url, endpoint=url))
        return self

    def map_routes(self, routes_map: dict):
        """
        注册路由映射表
        :param routes_map:
        :return:
        """
        for path in routes_map:
            if not path:
                raise Exception('Empty route map prefix value')
            pkg = routes_map[path]
            pkg_path = os.path.join(self.config.ROOT, pkg.replace('.', '/'))
            if not os.path.exists(pkg_path):
                self._logger.warning('Route map "%s" does not exist, path=%s' % (pkg, pkg_path))
            self.config.routes_map[path] = pkg
        return self

    def map_static(self, static_map: dict):
        """
        设置静态资源目录映射
        :param static_map:
        :return:
        """
        for prefix in static_map:
            # 静态资源的前缀，必须以 / 开头
            if prefix[0] != '/':
                self._logger.warning(
                    'Invalid prefix of static map, it should start with "/": %s' % prefix)
                continue
            if prefix.lower() == '/internal_assets':
                self._logger.warning(
                    'The prefix of static map is reserved, pick another one please: %s' % prefix)
                continue
            target_path = static_map[prefix]
            abs_target_path = os.path.abspath(os.path.join(self.config.ROOT, target_path))
            # 无论目录是否存在都注册，因为静态目录可能是动态生成的
            self.config.static_map[prefix] = abs_target_path
            if os.path.exists(abs_target_path):
                self._logger.debug('Map static url %s to path %s' % (prefix, abs_target_path))
            else:
                self._logger.warning(
                    'The target path of static url %s not found: %s' % (prefix, abs_target_path))
        return self

    def register_routes(self, routes: list):
        """
        手动注册路由列表
        :param routes: 其每一项都应该是一个 list, 元素依次为 method: str, path: str, handler: FunctionType
        :return:
        """
        for route in routes:
            self.register_route(*route)
        return self

    def register_route(self, method: str, path: str, handler: FunctionType):
        """
        手动注册路由
        :param path:
        :param method:
        :param handler:
        :return:
        """
        from .util.func_util import FunctionDescription

        rid = '%s#%s' % (path, method.lower())
        if rid in self._router.production_routes:
            self._logger.warning('Duplicated route %s %s' % (method, path))

        desc = FunctionDescription(handler)
        self._router.production_routes[rid] = {
            'func': handler,
            'args': desc.arguments
        }
        return self

    def register_middleware(self, *middlewares):
        """
        注册中间件，注册的中间件将按顺序执行
        :param middlewares: 中间件实例列表
        :return:
        """
        from restfx.middleware import MiddlewareBase

        for middleware in middlewares:
            assert isinstance(middleware, MiddlewareBase)
            self.config.middlewares.append(middleware)
            self.config.reversed_middlewares.insert(0, middleware)
            # 调用中间件，标记中间件的启动
            middleware.on_startup(self)

        return self

    def inject(self, **kwargs):
        self.config.injections.update(**kwargs)
        return self

    @classmethod
    def get(cls, app_id: str):
        """
        获取指定的应用
        :param app_id:
        :return:
        :rtype: App
        """
        return cls._APPS.get(app_id)


@atexit.register
def _destroy_apps():
    # Parse the ids into list to avoid the error "RuntimeError: dictionary changed size during iteration"
    app_ids = list(App._APPS.keys())
    for app_id in app_ids:
        app = App.get(app_id)
        app.dispose()



import os
import sys
import uuid

from restfx import __meta__
from restfx.util import helper

_COMMANDS = {}


def register(command_name: str, handler, description: str = None, args: str = None):
    _COMMANDS[command_name] = {
        'name': command_name,
        'handler': handler,
        'arguments': args,
        'description': '' if description is None else description
    }


def run_command(working_dir: str, command_name: str, *args):
    helper.print_meta(command_name)

    command = get_command(command_name)
    if command is None:
        return
    try:
        command['handler'](working_dir, *args)
    except Exception as e:
        raise e


def get_command(command_name: str):
    if command_name not in _COMMANDS:
        print('Command %s not found' % command_name)
        return None
    return _COMMANDS[command_name]


def get_commands():
    return [
        '%s%s\t%s' % (
            command['name'],
            (' <%s>' % command['arguments']) if command['arguments'] else '',
            command['description']
        )
        for command in _COMMANDS.values()
    ]


# noinspection PyUnusedLocal
def command_help(*argv):
    commands = get_commands()
    print("""Usage: {name} <command> [arguments]

Commands:
\t{commands}

Documentation links:
    https://gitee.com/hyjiacan/restfx
    https://github.com/hyjiacan/restfx
""".format(name=__meta__.name,
           commands='\n\t'.join(commands)))


# noinspection PyUnusedLocal
def command_version(*argv):
    print('Version: ' + __meta__.version)


# noinspection PyUnusedLocal
def command_create(working_dir: str, project_name, *argv):
    print('working-dir:' + working_dir)
    pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    project_path = os.path.abspath(os.path.join(working_dir, project_name))

    print('Creating restfx project "%s"' % project_name)

    if os.path.exists(project_path):
        print('[ERROR] Project path "%s" exists.' % project_path)
        sys.exit(1)

    # 示例文件
    sample_file = os.path.abspath(os.path.join(pkg_root, 'internal_assets', 'sample.tar.gz'))
    print('Creating project structure')
    import tarfile
    tarfile = tarfile.open(sample_file)
    tarfile.extractall(project_path)
    tarfile.close()

    print('Generating APP ID')
    # 生成 uuid
    setting_file = os.path.abspath(os.path.join(project_path, 'settings.py'))
    with open(setting_file, mode='r', encoding='utf8') as fp:
        lines = fp.readlines()
        app_id = uuid.uuid4()
        print(app_id)
        # 在第5行
        lines[5] = "APP_ID = '%s'%s" % (app_id, os.linesep)

    with open(setting_file, mode='w', encoding='utf8') as fp:
        fp.writelines(lines)

    print("""Created !

Now you can enjoy restfx.
""".format(project_name=project_name))



# a singleton sentinel value for parameter defaults
_sentinel = object()


class ContextStore:
    """
    上下文中数据存储支持
    """

    def __init__(self, ctx_stack):
        self.store = {}
        self.ctx_stack = ctx_stack

    def get(self, name, default=None):
        return self.store.get(name, default)

    def pop(self, name, default=_sentinel):
        if default is _sentinel:
            return self.store.pop(name)
        else:
            return self.store.pop(name, default)

    def setdefault(self, name, default=None):
        return self.store.setdefault(name, default)

    def update(self, items: dict):
        self.store.update(items)

    def __contains__(self, item):
        return item in self.store

    def __iter__(self):
        return iter(self.store)

    def __repr__(self):
        top = self.ctx_stack.top
        if top is not None:
            return "<restfx.ContextStore of %r>" % top.app.id
        return object.__repr__(self)



from ..interface import MiddlewareBase


class HttpAuthMiddleware(MiddlewareBase):
    def __init__(self, on_auth, auth_type='basic', realm='.'):
        self.on_auth = on_auth
        self.auth_type = auth_type
        self.realm = realm

    def process_request(self, request, meta):
        if self.on_auth(request, meta):
            return
        from restfx.http.response import Unauthorized
        response = Unauthorized()
        response.headers.set('WWW-Authenticate', '%s realm="%s"' % (self.auth_type, self.realm))
        return response



from types import FunctionType

from .route_resolver import RouteResolver
from ..config import AppConfig
from ..http import HttpResponse, NotFound
from ..http.request import HttpRequest


class Router:
    def __init__(self, config: AppConfig):
        self.config = config
        # 函数缓存，减少 inspect 反射调用次数
        self.entry_cache = {}
        """
        :type:Dict[str, Optional[FunctionDescription]]
        """
        # 线上模式时，使用固定路由
        self.production_routes = {}

    def dispatch(self, request: HttpRequest, entry):
        """
        路由分发入口
        :param request: 请求
        :param entry: 入口地址
        :return:
        """
        if not self.config.debug:
            return self.route_for_production(request, '/' + entry)

        resolver = RouteResolver(self.config,
                                 self.entry_cache,
                                 request.method, entry)

        route = resolver.resolve()
        if isinstance(route, HttpResponse):
            return route

        return self.invoke_handler(request, route.func, route.arguments)

    def route_for_production(self, request, entry):
        method = request.method.lower()
        # noinspection PyBroadException
        try:
            route = self.production_routes['%s#%s' % (entry, method)]
        except Exception:
            return NotFound()

        return self.invoke_handler(request, route['func'], route['args'])

    def invoke_handler(self, request, func: FunctionType, args):
        try:
            return func(request, args)
        except Exception as e:
            from ..util import Logger, func_util
            Logger.get(self.config.app_id).error(func_util.get_func_info(func), e)
            raise e



from restfx import __meta__


def print_meta(append=None):
    info = '{name} {version}'.format(name=__meta__.name, version=__meta__.version)
    if append is not None:
        info += '  <%s>' % append

    project_url = __meta__.website

    info_len = len(info)
    url_len = len(project_url)

    if info_len > url_len:
        max_len = info_len
        project_url = project_url.center(info_len, ' ')
    else:
        max_len = url_len
        info = info.center(url_len, ' ')

    line_width = '-' * (max_len + 8)
    print("#{line}#\n|    {info}    |\n|    {url}    |\n#{line}#\n".format(
        line=line_width, info=info, url=project_url))



import os

from werkzeug.local import LocalStack

from restfx import __meta__


class AppConfig:
    """
    应用的配置
    """

    # 所有应用的配置集合
    _CONFIGS = {}

    def __init__(self, app_id: str,
                 app_root: str,
                 debug: bool,
                 append_slash: bool,
                 strict_mode: bool,
                 api_page_enabled: bool,
                 api_page_name: str,
                 api_page_expanded: bool,
                 api_page_cache: bool,
                 api_page_addition,
                 api_page_header: str,
                 api_page_footer: str,
                 api_page_assets: tuple,
                 ):
        """

        """
        self._CONFIGS[app_id] = self
        self.app_id = app_id
        self.local = LocalStack()
        # 是否启用DEBUG模式
        self.debug = debug
        # 工作目录
        self.ROOT = app_root
        self.append_slash = append_slash
        self.strict_mode = strict_mode
        self.middleware_manager = None
        # 注册的中间件实例集合
        self.middlewares = []
        """
        :type: List[MiddlewareBase]
        """
        self.reversed_middlewares = []
        """
        :type: List[MiddlewareBase]
        """

        # 路由映射表，其键为请求的路径，其值为映射的目录
        self.routes_map = {}

        self.static_map = {}

        # 注入数据/函数集合
        # 其中存放将被注入到路由函数参数列表上的数据/函数
        self.injections = {}

        # 是否启用 API 页面
        self.api_page_enabled = debug if api_page_enabled is None else api_page_enabled
        self.api_page_name = api_page_name or 'An awesome %s project' % __meta__.name
        self.api_page_expanded = api_page_expanded
        # 是否缓存API页面的 html 文件 和 接口数据
        self.api_page_cache = not debug if api_page_cache is None else api_page_cache
        self.api_page_header = api_page_header
        self.api_page_footer = api_page_footer
        self.api_page_addition = api_page_addition
        self.api_page_assets = api_page_assets

        if self.api_page_enabled:
            self.static_map['/internal_assets'] = os.path.join(os.path.dirname(__file__), 'internal_assets')

    def __del__(self):
        if self.app_id in self._CONFIGS:
            del self._CONFIGS[self.app_id]
        del self.middlewares
        del self.reversed_middlewares

    @classmethod
    def get(cls, app_id: str):
        """

        :param app_id:
        :return:
        :rtype: AppConfig
        """
        return cls._CONFIGS.get(app_id)



import pickle
import time
from abc import ABC, abstractmethod
from threading import Timer
from typing import List, Optional

from .session import HttpSession


class ISessionProvider(ABC):
    """
    session 提供者基类，自定义 session 提供者时应该继承此类
    此类不可被实例化
    """

    def __init__(self, expired: int):
        """

        :param expired: 过期时长，单位为秒，默认为 10分钟。指定为 0表示不会自动过期
        """
        self.expired = 10 * 60 if expired is None else expired

        # 不会过期，不用启动定时器
        if self.expired > 0:
            # 每 5 秒执行一次检查
            self.timer = Timer(5, self._drop_expire_session)
            self.timer.start()

    def _drop_expire_session(self):
        time_before = time.time() - self.expired
        expired_sessions = self.get_expired_session(time_before)
        for session_id in expired_sessions:
            # print('Drop expired session:' + session_id)
            self.remove(session_id)

    def is_expired(self, session: HttpSession) -> bool:
        if self.expired > 0:
            return time.time() - session.last_access_time > self.expired
        return False

    def create(self, session_id: str) -> HttpSession:
        session = HttpSession(session_id, self.set, self.remove)
        # 实际存储了数据再创建
        # self.set(session)
        return session

    @abstractmethod
    def get_expired_session(self, time_before: float) -> List[str]:
        """
        查询已经过期的 session 的 id
        :param time_before: 在此时间之前的 session 即为过期
        :return:
        """
        pass

    @abstractmethod
    def get(self, session_id: str) -> Optional[HttpSession]:
        """
        获取指定 id 的 session
        :param session_id:
        :return:
        """
        pass

    @abstractmethod
    def set(self, session: HttpSession):
        """
        添加或更新指定的 session
        :param session:
        :return:
        """
        pass

    @abstractmethod
    def exists(self, session_id: str) -> bool:
        """
        检索指定 id 的 session 是否存在
        :param session_id:
        :return:
        """
        pass

    @abstractmethod
    def remove(self, session_id: str):
        """
        销毁一个 session
        :param session_id:
        :return:
        """

    @abstractmethod
    def dispose(self):
        """
        回收 session provider
        :return:
        """
        self.timer.cancel()
        self.timer.join()


class IDbSessionProvider(ISessionProvider):
    def __init__(self, pool_options: dict, expired: int):
        from dbutils.pooled_db import PooledDB
        self.pool_option = pool_options
        self.pool = PooledDB(**pool_options)
        # fixme 如果启动时数据库无法连接，也应该允许程序启动，而不是抛出异常
        # 更好的做法是在第一次执行 execute 时作此检查
        if not self.table_exists():
            self.create_table()
        super().__init__(expired)

    def connect(self, shareable=True):
        return self.pool.connection(shareable)

    @abstractmethod
    def table_exists(self) -> bool:
        """
        判断 session 表是否存在
        :return:
        """
        pass

    @abstractmethod
    def create_table(self):
        """
        创建 session 表
        :return:
        """
        pass

    def parse(self, data: dict) -> HttpSession:
        """
        用于将字典转换成 HttpSession 对象
        :param data:
        :return:
        """
        session = HttpSession(data['id'], self.set, self.remove)
        session.creation_time = int(data['creation_time'])
        session.last_access_time = int(data['last_access_time'])
        session.store = pickle.loads(data['store'])

        return session

    def set(self, session: HttpSession):
        self.upsert(session.id,
                    session.creation_time,
                    session.last_access_time,
                    pickle.dumps(session.store))

    @abstractmethod
    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        pass

    @abstractmethod
    def dispose(self):
        self.pool.close()
        super(IDbSessionProvider, self).dispose()



import os
import sys

from .commands import (
    run_command,
    register,
    command_help,
    command_create,
    command_version
)

register('help', command_help, 'Show this help message')
register('version', command_version, 'Show the version information')
register('create', command_create, 'Create restfx project structure', 'project-name')


def execute(argv=None):
    args = (argv or sys.argv)[1:]

    working_dir = os.path.abspath(os.path.curdir)

    if len(args) == 0:
        run_command(working_dir, 'help')
        return

    command_name = args.pop(0)
    run_command(working_dir, command_name, *args)



import os

from .app import App
from .routes.decorator import route
from .routes.parameter_interface import IParam
from .routes.validator import Validator

val = Validator
"""
Validator 的简短别名
"""

env = os.environ.get('RESTFX_ENV')
"""
指定执行环境环境，通常情况下，可选值为 dev 或者 prod(默认)
"""

# 如果没有指定，默认使用 prod
if not env:
    env = 'prod'

__all__ = [
    'App',
    'route',
    'val',
    'IParam',
    'env'
]



import json
from io import IOBase
from typing import Tuple, Union

from werkzeug.exceptions import InternalServerError
from werkzeug.utils import escape
from werkzeug.wrappers import Response


class HttpResponse(Response):
    def __init__(self, content=None, content_type='text/html;charset=utf-8',
                 status=200, headers=None, **kwargs):
        super().__init__(content, content_type=content_type, status=status,
                         headers=headers, **kwargs)


class JsonResponse(HttpResponse):
    def __init__(self, obj, encoder=None, content_type='application/json;charset=utf-8',
                 ensure_ascii=True, **kwargs):
        content = json.dumps(obj, ensure_ascii=ensure_ascii, cls=encoder)
        super().__init__(content, content_type=content_type, **kwargs)


class FileResponse(HttpResponse):
    def __init__(self, fp: Union[str, bytes, IOBase], attachment: str = None,
                 content_type=None,
                 ranges: Tuple[int, int] = (),
                 request=None, **kwargs):
        """

        :param fp: 文件名或内容。如果指定的 fp 是文件名（字符串），那么认为返回的是文件名。
        如果需要返回的内容是字符串，那么需要先处理成 bytes
        :param attachment: 指定一个字符串，作为返回附件的文件名
        :param content_type: 当未指定此值时，如果指定的 fp 是文件名，那么会自动根据文件的扩展名进行识别
        :param ranges: 用于指定返回数据的分块起始位置
        :param kwargs:
        """
        # 如果是字符串，就认为是文件路径
        if isinstance(fp, str):
            # 根据文件的扩展名自动识别 mime
            if content_type is None:
                import mimetypes
                (content_type, _) = mimetypes.guess_type(fp)
            self.fp = open(fp, mode='rb')
        else:
            self.fp = fp

        if content_type is None:
            content_type = 'application/octet-stream'

        # 需要分块返回文件
        if ranges:
            # fix https://gitee.com/hyjiacan/restfx/issues/I3SBXR
            # IOBase for BufferedReader(with open(xxx)) and BytesIO, etc.
            if not isinstance(self.fp, IOBase):
                raise TypeError('FileResponse with "ranges" works with type "IOBase" only')
            if not self._get_file_chunk(ranges, kwargs):
                super(FileResponse, self).__init__(status=416)
                return
            status_code = 206
        else:
            status_code = 200

        if attachment:
            self._set_attachment_header(request, attachment, kwargs)

        super().__init__(self.fp, status=status_code, direct_passthrough=True,
                         content_type=content_type, **kwargs)

    def _get_file_chunk(self, ranges: Tuple[int, int], kwargs):
        from io import BytesIO

        pos = self.fp.tell()
        self.fp.seek(0, 2)
        file_size = self.fp.tell()
        self.fp.seek(pos, 0)

        start, end = ranges

        if start >= file_size or end >= file_size:
            return False

        if end == 0:
            # to the end
            self.fp.seek(start)
            chunk_size = file_size - start
        else:
            chunk = BytesIO()
            # The fetched data size
            chunk_size = end - start
            self.fp.seek(start)
            buffer = self.fp.read(chunk_size)
            self.fp.close()
            chunk.write(buffer)
            chunk.seek(0)
            self.fp = chunk

        # set the response headers
        headers = kwargs.get('headers')
        if headers is None:
            headers = {}
            kwargs['headers'] = headers

        headers['Accept-Ranges'] = 'bytes'
        headers['Content-Length'] = str(chunk_size)
        headers['Content-Range'] = 'bytes %s-%s/%s' % (start, end, file_size)
        return True

    # noinspection PyMethodMayBeStatic
    def _set_attachment_header(self, request, filename, kwargs):
        from urllib.parse import unquote, quote

        headers = kwargs.get('headers')
        if headers is None:
            headers = {}
            kwargs['headers'] = headers

        if not request:
            from ..util import Logger
            Logger.print('warning',
                         'You are using FileResponse for attachment,'
                         'it is recommended to fill the "request" parameter.'
                         'Otherwise, you may got an encoding issue of the filename on Firefox.'
                         )

        user_agent = request and request.headers.environ['HTTP_USER_AGENT']
        is_firefox = user_agent and ('Firefox/' in user_agent)

        if is_firefox:
            from ..util import b64
            filename = '=?utf-8?B?' + b64.enc_bytes(unquote(filename).encode('utf-8')).decode() + '?='
        else:
            filename = quote(filename)

        headers['Content-Disposition'] = 'attachment;filename=%s' % filename


class BadRequest(HttpResponse):
    def __init__(self, content=None, **kwargs):
        super().__init__(content, status=400, **kwargs)


class Redirect(HttpResponse):
    def __init__(self, location, status=302, **kwargs):
        super().__init__(None, status=status, headers={
            'Location': escape(location)
        }, **kwargs)


class Unauthorized(HttpResponse):
    def __init__(self, content=None, **kwargs):
        super().__init__(content, status=401, **kwargs)


class NotFound(HttpResponse):
    def __init__(self, content=None, **kwargs):
        super().__init__(content, status=404, **kwargs)


class ServerError(HttpResponse, InternalServerError):
    def __init__(self, content=None):
        if isinstance(content, Exception):
            content = content.args[0] if content.args else str(content)
        super().__init__(content, status=500, content_type='text/plain')



from .request import HttpRequest, HttpFile
from .response import (
    HttpResponse,
    ServerError,
    BadRequest,
    NotFound,
    JsonResponse,
    FileResponse,
    Redirect, Unauthorized
)

__all__ = [
    'HttpFile',
    'HttpRequest',
    'HttpResponse',
    'FileResponse',
    'JsonResponse',
    'Redirect',
    'BadRequest',
    'Unauthorized',
    'NotFound',
    'ServerError'
]



from types import FunctionType


class RouteMeta:
    """
    路由元数据
    """

    def __init__(self,
                 handler: FunctionType,
                 handler_args,
                 route_id=None,
                 module=None,
                 name=None,
                 extname=None,
                 method=None,
                 path=None,
                 kwargs=None):
        """

        :param handler: 路由函数对象
        :param handler_args: 路由函数参数列表
        :param route_id: 路由ID，此ID由路由相关信息组合而成
        :param module: 装饰器上指定的 module 值
        :param name: 装饰器上指定的 name 值
        :param extname: 装饰器上指定的 name 值
        :param kwargs: 装饰器上指定的其它参数
        """
        self.handler = handler
        self.handler_args = handler_args
        self.id = route_id
        self.module = module
        self.name = name
        self.extname = extname
        self.method = method
        self.path = path
        self.kwargs = {} if kwargs is None else kwargs

    def has(self, arg_name):
        """
        指定的参数是否存在
        :param arg_name:
        :return:
        """
        return arg_name in self.kwargs

    def get(self, arg_name: str, default_value=None):
        """
        获取指定参数的值
        :param default_value:
        :param arg_name:
        :return:
        """
        return self.kwargs.get(arg_name, default_value)



import uuid
from werkzeug import Request
from werkzeug.datastructures import ImmutableDict, ImmutableMultiDict, FileStorage

from restfx.util import ContextStore
from restfx.globals import _request_ctx_stack, _app_ctx_stack


def _get_request_data(data: ImmutableMultiDict) -> ImmutableDict:
    args = {}
    for key in data.keys():
        value = data.getlist(key)
        value_len = len(value)
        if not value_len:
            args[key] = None
            continue
        if value_len == 1:
            args[key] = value[0]
            continue
        args[key] = value
    return ImmutableDict(args)


def _get_files(files):
    if not files:
        return None
    result = {}
    for name in files:
        file = files.get(name)
        result[name] = HttpFile.inherit(file)

    return ImmutableMultiDict(result)


class RequestContext:
    def __init__(self, request):
        self.request = request
        self.store = ContextStore(_request_ctx_stack)
        self.ref_count = 0

    @property
    def session(self):
        return self.request.session

    def push(self):
        top = _app_ctx_stack.top
        if not top or top != self.request.app:
            self.request.app.context.push()
        if _request_ctx_stack.top != self:
            _request_ctx_stack.push(self)
        self.ref_count += 1

    def pop(self):
        if _request_ctx_stack.top != self:
            return
        self.ref_count -= 1
        if self.ref_count > 0:
            return
        _request_ctx_stack.pop()
        del self.store

    def __enter__(self):
        self.push()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pop()


class HttpRequest(Request):
    def __init__(self, environ, app):
        super().__init__(environ)
        self.id = str(uuid.uuid4())
        self.app = app
        self.app_id = app.id
        self.GET = _get_request_data(self.args)
        self.POST = _get_request_data(self.form)
        self.BODY = self.data
        self.FILES = _get_files(self.files)
        self.COOKIES = self.cookies
        self.session = None
        self._injections = {}
        # 用于存放一些用户的数据
        self._user_data = {}
        # 在同一个请求中，使用同一个请求上下文
        self._ctx = RequestContext(self)

    def inject(self, **kwargs):
        self._injections.update(kwargs)

    def context(self, **kwargs):
        self._ctx.store.update(kwargs)
        return self._ctx

    def get(self, key: str, default=None):
        """
        读取附加到 request 上的自定义数据
        :param key:
        :param default:
        :return:
        """
        return self._user_data.get(key, default)

    def set(self, key: str, value):
        """
        设置附加到 request 上的自定义数据
        :param key:
        :param value:
        :return:
        """
        return self._user_data.setdefault(key, value)

    def remove(self, key: str):
        """
        移除附加到 request 上的自定义数据
        :param key:
        :return:
        """
        return self._user_data.pop(key)

    @property
    def range(self):
        import re
        request_range = self.headers.get('range')
        if not request_range:
            return None
        match = re.fullmatch(r'\s*(?P<unit>(bytes))=(?P<start>([0-9]+))-(?P<end>([0-9]*))?\s*', request_range)
        if not match:
            from restfx.util import Logger
            from restfx.http import BadRequest
            Logger.get(self.app_id).warning('Invalid value of request header "range": %r' % request_range)
            return BadRequest()

        return match.group('unit'), int(match.group('start')), int(match.group('end') or 0)


class HttpFile(FileStorage):
    @property
    def size(self):
        """
        此功能处于试用阶段，尚未知晓其可能产生的副作用
        :return:
        """
        # noinspection PyBroadException
        try:
            position = self.stream.tell()
            self.stream.seek(0, 2)
            size = self.stream.tell()
            self.stream.seek(position, 0)
            return size
        except Exception:
            return self.content_length

    @staticmethod
    def inherit(file: FileStorage):
        return HttpFile(
            stream=file.stream,
            filename=file.filename,
            name=file.name,
            content_type=file.content_type,
            content_length=file.content_length,
            headers=file.headers
        )



import time

from restfx import __meta__
from restfx.middleware import MiddlewareBase


class TimetickMiddleware(MiddlewareBase):
    route_time = __meta__.name + '-duration-0-route'
    process_time = __meta__.name + '-duration-1-process'
    invoke_time = __meta__.name + '-duration-2-invoke'
    response_time = __meta__.name + '-duration-3-response'

    def __init__(self):
        pass

    def on_coming(self, request):
        request.timetick = {
            'request': time.perf_counter()
        }

    def process_request(self, request, meta):
        if request.timetick:
            request.timetick['dispatch'] = time.perf_counter()

    def process_invoke(self, request, meta, args):
        if request.timetick:
            request.timetick['invoke'] = time.perf_counter()

    def process_return(self, request, meta, data):
        if request.timetick:
            request.timetick['return'] = time.perf_counter()

    def process_response(self, request, meta, response):
        if not request.timetick:
            return

        request.timetick['response'] = time.perf_counter()

        rd = self._get_time(request, 'request', 'dispatch')
        dt = self._get_time(request, 'dispatch', 'invoke')
        it = self._get_time(request, 'invoke', 'return')
        rpt = self._get_time(request, 'return', 'response')

        print('%s=%s\n%s=%s\n%s=%s\n%s=%s\n' % (
            self.route_time, rd,
            self.process_time, dt,
            self.invoke_time, it,
            self.response_time, rpt
        ))

        response.headers.set(self.route_time, rd)
        response.headers.set(self.process_time, dt)
        response.headers.set(self.invoke_time, it)
        response.headers.set(self.response_time, rpt)

    @staticmethod
    def _get_time(request, from_, to):
        if from_ not in request.timetick or to not in request.timetick:
            return '0ms'
        return '%sms' % ((request.timetick[to] - request.timetick[from_]) * 1000)



from abc import ABC, abstractmethod


class IParam(ABC):
    """
    自定义参数类型时，通过实现此接口以进行参数类型的转换
    """

    @classmethod
    @abstractmethod
    def parse(cls, value: (list, dict)):
        raise NotImplemented()



from abc import ABC


class MiddlewareBase(ABC):
    """
    路由中间件基类
    """

    def on_startup(self, app):
        """
        在注册中间件后立即调用
        :param app: App 实例
        :return: 不需要返回任何值
        """
        pass

    def on_shutdown(self):
        """
        在应用停止时调用
        :return:
        """
        pass

    def on_coming(self, request):
        """
        在收到请求时立即调用
        :param request:
        :return:
        返回 HttpResponse 以终止请求，
        返回 None 继续执行后续中间件，
        返回其它类型数据将作为请求的返回数据（会跳过路由调用）
        """
        pass

    def on_leaving(self, request, response):
        """
        在请求线束时调用，无论是否发生了异常
        :param request:
        :param response:
        :return:
        返回 HttpResponse 以终止请求，
        返回 None 继续执行后续中间件，
        返回其它类型数据将作为请求的返回数据（会跳过路由调用）
        """
        pass

    def process_request(self, request, meta):
        """
        对 request 对象进行预处理。一般用于请求的数据的解码，此时路由组件已经完成了请求数据的解析(BODY,POST,GET 已可用)
        :param request:
        :type request: HttpRequest
        :param meta:
        :type meta: RouteMeta
        :return:
        返回 HttpResponse 以终止请求，
        返回 None 继续执行后续中间件，
        返回其它类型数据将作为请求的返回数据（会跳过路由调用）
        """
        pass

    def process_invoke(self, request, meta, args: dict):
        """
        在路由函数调用前，对传递给路由函数参数等进行处理
        此时可以对解析后的参数进行变更
        :param request:
        :type request: HttpRequest
        :param meta:
        :type meta: RouteMeta
        :param args: 调用路由函数时的实际参数，可以通过修改此参数来改变传给路由函数的参数
        :type args: dict
        :return:
        返回 HttpResponse 以终止请求，
        返回 None 继续执行后续中间件，
        返回其它类型数据将作为请求的返回数据（会跳过路由调用）
        """
        pass

    def process_return(self, request, meta, data):
        """
        在路由函数调用后，对其返回值进行处理
        :param request:
        :type request: HttpRequest
        :param meta:
        :type meta: RouteMeta
        :param data: 表示路由返回的原始数据
        :return:
        返回 HttpResponse 以终止请求，
        返回非 None 时，其值作为下一个中间件的输入 data，
        返回 None 继续执行后续中间件
        """
        pass

    def process_response(self, request, meta, response):
        """
        对 response 数据进行预处理。一般用于响应的数据的编码
        :param request:
        :type request: HttpRequest
        :param meta:
        :type meta: RouteMeta
        :param response: 表示路由返回的原始 HttpResponse
        :type response: HttpResponse
        :return:
        返回 HttpResponse 类型时，其值作为下一个中间件的输入 response，
        返回 None 继续执行后续中间件
        """
        pass



from ..http.response import HttpResponse
from ..util import Logger, utils


class MiddlewareManager:
    """
    路由中间件管理器
    """

    def __init__(self, app_id: str, config):
        self.config = config
        self.logger = Logger.get(app_id)

    def handle_shutdown(self):
        for middleware in self.config.reversed_middlewares:
            try:
                middleware.on_shutdown()
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.on_shutdown), e)
                raise e

    def handle_coming(self, request):
        for middleware in self.config.middlewares:
            try:
                result = middleware.on_coming(request)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.on_coming), e)
                raise e
            # 返回的值为 None，继续执行下一个中间件
            if result is None:
                continue
            # 返回其它类型，中止请求
            return result

    def handle_leaving(self, request, response):
        for middleware in self.config.reversed_middlewares:
            try:
                result = middleware.on_leaving(request, response)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.on_leaving), e)
                raise e
            # 返回的值为 None，继续执行下一个中间件
            if result is None:
                continue
            # 返回其它类型，中止请求
            return result

    def handle_request(self, request, meta):
        for middleware in self.config.middlewares:
            try:
                result = middleware.process_request(request, meta)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.process_request), e)
                raise e
            # 返回的值为 None，继续执行下一个中间件
            if result is None:
                continue
            # 返回其它类型，中止请求
            return result

    def before_invoke(self, request, meta, args: dict):
        """
        在路由函数调用前，对其参数等进行处理
        :return:
        """
        for middleware in self.config.middlewares:
            try:
                result = middleware.process_invoke(request, meta, args)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.process_invoke), e)
                raise e
            # 返回的值为 None，继续执行下一个中间件
            if result is None:
                continue
            # 返回其它类型，中止请求
            return result

    def after_return(self, request, meta, data):
        """
        在路由函数调用后，对其返回值进行处理
        :return:
        """
        for middleware in self.config.reversed_middlewares:
            try:
                result = middleware.process_return(request, meta, data=data)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.process_return), e)
                raise e
            # 返回的值为 None，继续执行下一个中间件
            if result is None:
                continue
            # 返回 HttpResponse，中止请求
            if isinstance(result, HttpResponse):
                return result
            # 使用返回值作为下一个中间件的数据
            data = result
        return data

    def handle_response(self, request, meta, response):
        """
        在响应前，对响应的数据进行处理
        :return:
        """
        # 对 response 进行处理
        for middleware in self.config.reversed_middlewares:
            try:
                result = middleware.process_response(request, meta, response=response)
            except Exception as e:
                self.logger.error(utils.get_func_info(middleware.process_response), e)
                raise e
            # 返回 None，那么继续执行下一个中间件
            if result is None:
                continue
            # 返回的是 HttpResponse，使用此值代替原来的值
            if isinstance(result, HttpResponse):
                response = result
                continue
            # 返回的不是 HttpResponse，返回的值类型不正确
            # 返回其它类型
            self.logger.warning('%s\n\tIgnored return type "%s", expect "None/HttpResponse"' % (
                utils.get_func_info(middleware.process_invoke), type(result).__name__
            ))

        return response



# coding: utf-8

import ast
# 此模块用于收集路由
import os
import re
from os import path

from ..util import utils
from ..util.func_util import FunctionDescription

# 生成：注册路由的代码 -- 模板
# 注意生成的代码中的缩进，使用的是空格

_REGISTER_STMT = "    # {module}-{name}\n    ['{method}', '{path}', {handler}]"
_CODE_TPL = """# -*- coding={encoding} -*-

# Note:
# This file is auto generated by restfx, DO NOT modify it manually.
# Any changes will be discarded once "app.persist" is called.

# IMPORT ROUTES BEGIN
{imports}
# IMPORT ROUTES END


# LIST ROUTES BEGIN
routes = [
{routes}
]
# LIST ROUTES END
"""


class Collector:
    _COLLECTORS = {}

    def __init__(self, app_id: str, app_root: str, append_slash: bool):
        self.app_id = app_id
        self.app_root = app_root
        self.append_slash = append_slash
        self.global_types = {}

    @classmethod
    def create(cls, app_id: str, app_root: str, append_slash: bool):
        collector = Collector(app_id, app_root, append_slash)
        cls._COLLECTORS[app_id] = collector
        return collector

    def collect(self, routes_map: dict):
        """
        执行收集操作
        :return: 所有路由的集合
        """
        # 所有路由的集合
        routes = []

        if not routes_map:
            raise Exception(
                'Routes map is empty, did you forgot to call "restfx.map_routes(routes_map: dict)"')

        for (http_prefix, pkg_prefix) in routes_map.items():
            route_root = path.abspath(path.join(self.app_root, pkg_prefix.replace('.', path.sep)))

            # 遍历目录，找出所有的 .py 文件
            for (dir_name, dirs, files) in os.walk(route_root):
                if dir_name == '__pycache__':
                    continue
                for file in files:
                    # 不是 .py 文件，忽略
                    if not file.endswith('.py'):
                        continue

                    # 可能是 __init__.py
                    fullname = path.abspath(path.join(dir_name, file))

                    # 解析文件
                    self.get_route_defines(route_root, fullname, http_prefix, pkg_prefix, routes)

        return routes

    def get_route_defines(self, route_root, fullname, http_prefix, pkg_prefix, routes):
        for define in self.resolve_file(route_root, fullname, http_prefix, pkg_prefix):
            # 返回 None 表示没有找到定义
            if define is None:
                continue
            routes.append(define)

    def resolve_file(self, route_define, fullname, http_prefix, pkg_prefix):
        """
        解析文件
        :param pkg_prefix:
        :param http_prefix: http 请求前缀
        :param route_define: 路由文件的根路径
        :param fullname: 文件的完整路径
        :return: 没有路由时返回 None
        """
        # 解析路由的定义
        routes = self.resolve_routes(fullname)

        # 没有找到定义，返回 None
        if not routes:
            yield None

        # router_info 是在函数上声明的装饰器定义信息
        # func_name 是函数的名称
        for (func_name, router_info) in routes:
            # 解析出请求的方法(method)与请求的指定函数名称
            method, _, name = re.match(r'([a-z]+)(_(.+))?', func_name).groups()

            # 构造http请求的地址(将 路径分隔符号 \/ 替换成 . 符号)
            # -3 是为了干掉最后的 .py 字样
            pkg = re.sub(r'[/\\]', '.', path.relpath(fullname, route_define))[0:-3]

            is_package = path.basename(fullname) == '__init__.py'

            # 当是包时，移除 __init__ 部分
            if is_package:
                http_path = '%s.%s' % (http_prefix, pkg[0:-len('__init__')])
            else:
                http_path = '%s.%s' % (http_prefix, pkg)

            # 当 map_routes 指定的 http_prefix 为空时，前导的 . 符号是多余的
            # 当路由文件为根目录下的 __init__.py 时，没有可访问的文件名
            # 此时会出现得到的路由为  xxx. 的情况
            # 所以在此移除末尾FunctionDescription的 . 符号
            http_path = '/' + http_path.strip('.').replace('.', '/')

            # 如果指定了名称，就追加到地址后
            ext_mode = name is not None
            if ext_mode:
                http_path += '/' + name

            # 指定的 url 扩展名
            extname = router_info['extname']
            if extname is not None:
                http_path += '.' + extname

            if self.append_slash:
                http_path += '/'

            # 当是包时，移除 .__init__ 部分
            if is_package:
                pkg = '%s.%s' % (pkg_prefix, pkg[0:-len('__init__')])
            else:
                pkg = '%s.%s' % (pkg_prefix, pkg)

            pkg = pkg.rstrip('.')

            try:
                module = utils.load_module(pkg)
            except Exception as e:
                from ..util import Logger
                msg = 'File "%s", line %d, code %s\n\t%s' % (
                    fullname, router_info['lineno'], func_name,
                    'Cannot load module "%s"' % pkg
                )
                Logger.get(self.app_id).error(msg, e)
                raise e
            handler_obj = getattr(module, func_name)
            handler_info = FunctionDescription(handler_obj)

            # 唯一标识
            router_info['id'] = '%s_%s' % (pkg.replace('_', '__').replace('.', '_'), func_name)
            # 路由所在包名称
            router_info['pkg'] = pkg
            # 路由所在文件的完整路径
            router_info['file'] = fullname
            # 路由请求的处理函数
            router_info['handler'] = func_name
            # 路由的请求方法
            router_info['method'] = method
            # 路由的请求路径
            router_info['path'] = http_path
            # 路由函数的描述
            router_info['handler_info'] = handler_info
            # 是否是包
            router_info['is_package'] = is_package
            # 是否是扩展模式
            router_info['ext_mode'] = ext_mode
            # 自定义的装饰器参数
            router_info['kwargs'] = router_info['kwargs']

            yield router_info

    def persist(self, routes_map: dict, filename: str = '', encoding='utf8'):
        """
        将路由持久化
        :param routes_map:
        :param filename:
        :param encoding:
        :return: 持久化的 python 代码
        :rtype: str or None
        """
        imports = []
        routes = []

        print('Generating routes map...')
        for route in self.collect(routes_map):
            imports.append('from %s import %s as %s' % (route['pkg'], route['handler'], route['id']))
            routes.append(_REGISTER_STMT.format(
                module=route['module'],
                name=route['name'],
                method=route['method'].upper(),
                path=route['path'],
                handler=route['id']
            ))

        content = _CODE_TPL.format(encoding=encoding, imports='\n'.join(imports), routes=',\n'.join(routes))

        print('Routes map data generated')
        if not filename:
            return content

        print('Persisting into file %s file with encoding %s' % (filename, encoding))
        with open(filename, mode='wt', encoding=encoding) as fp:
            fp.write(content)
            fp.close()
        print('Routes persisted')

    def resolve_routes(self, filename: str, func_name: str = None):
        """

        :param filename:
        :param func_name: 解析指定的函数
        :return:
        """
        with open(filename, encoding='utf-8') as python_fp:
            lines = python_fp.readlines()
            python_fp.close()
        ast_body = ast.parse(source=''.join(lines), filename=filename).body

        routes = []

        for item in ast_body:
            if not isinstance(item, ast.FunctionDef):
                continue

            if func_name is not None and item.name != func_name:
                continue

            # Find out the @route decorator
            decorator_info = self.get_route_decorator(filename, item)
            if decorator_info is None:
                continue

            if func_name is not None:
                return decorator_info

            routes.append((item.name, decorator_info))

        return routes if func_name is None else None

    def get_route_decorator(self, filename: str, func_def: ast.FunctionDef):
        from ..util import Logger
        logger = Logger.get(self.app_id)
        for decorator in func_def.decorator_list:
            route_module = None
            route_name = None
            route_extname = None
            keywords = {}

            if hasattr(decorator, 'func'):
                if decorator.func.id != 'route':
                    continue
                for keyword in decorator.keywords:
                    arg_name = keyword.arg
                    value = keyword.value
                    if isinstance(value, ast.Attribute):
                        type_name = value.value.id
                        type_def = self.global_types[type_name]
                        type_val = value.attr
                        if not hasattr(type_def, type_val):
                            msg = 'File "%s", line %d, code %s\n\t%s' % (
                                filename, func_def.lineno, func_def.name,
                                'Cannot retrieve value "%s.%s" from type "%s"' % (type_name, type_val, type_name)
                            )
                            logger.error(msg)
                            raise Exception(msg)
                        arg_value = getattr(type_def, type_val)
                    else:
                        # 其它类型暂时不支持
                        # 统一使用原始值
                        # noinspection PyProtectedMember
                        arg_value = getattr(value, keyword.value._fields[0])

                    keywords[arg_name] = arg_value

                # 在此不需要考虑其它的数据类型，因为声明的时候全是 字符串
                if len(decorator.args) == 1:
                    route_module = decorator.args[0].s
                elif len(decorator.args) == 2:
                    route_module = decorator.args[0].s
                    route_name = decorator.args[1].s
                elif len(decorator.args) == 3:
                    route_module = decorator.args[0].s
                    route_name = decorator.args[1].s
                    route_extname = decorator.args[2].s

                if route_module is None and 'module' in keywords:
                    route_module = keywords['module']
                    del keywords['module']
                if route_name is None and 'name' in keywords:
                    route_name = keywords['name']
                    del keywords['name']
                if route_extname is None and 'extname' in keywords:
                    route_extname = keywords['extname']
                    del keywords['extname']
            elif not hasattr(decorator, 'id'):
                continue
            elif decorator.id != 'route':
                continue
            else:
                # 写了使用 @route 的方式（不带括号）
                # 这样的写法是错误的，应该写为  @route()
                msg = 'File "%s", line %d, code %s\n\t%s' % (
                    filename, func_def.lineno, func_def.name,
                    'Unexpected usage with "@route", use "@route()" instead.')
                raise Exception(msg)

            return {
                'lineno': func_def.lineno,
                'module': route_module,
                'name': route_name,
                'extname': route_extname,
                'kwargs': keywords
            }

        return None

    @classmethod
    def destroy(cls, app_id: str):
        if app_id in cls._COLLECTORS:
            del cls._COLLECTORS[app_id]

    @classmethod
    def get(cls, app_id: str):
        return cls._COLLECTORS.get(app_id, None)



import hashlib


def hash_bytes(s: (str, bytes)) -> bytes:
    if isinstance(s, str):
        s = s.encode(encoding='utf8')
    return hashlib.sha1(s).digest()


def hash_str(s: (str, bytes)) -> str:
    if isinstance(s, str):
        s = s.encode(encoding='utf8')
    return hashlib.sha1(s).hexdigest()



import os
from typing import List, Optional

from .interfaces import ISessionProvider, IDbSessionProvider
from .session import HttpSession


class MemorySessionProvider(ISessionProvider):
    def __init__(self, expired: int = None):
        self.sessions = {}
        super().__init__(expired)

    def remove(self, session_id: str):
        if self.exists(session_id):
            del self.sessions[session_id]

    def get_expired_session(self, time_before: float) -> List[str]:
        expired_sessions = []
        for session in self.sessions.values():
            if time_before > session.last_access_time:
                expired_sessions.append(session.id)
        return expired_sessions

    def get(self, session_id: str) -> Optional[HttpSession]:
        return self.sessions[session_id] if self.exists(session_id) else None

    def set(self, session: HttpSession):
        self.sessions[session.id] = session

    def exists(self, session_id: str):
        return session_id in self.sessions

    def dispose(self):
        self.sessions.clear()
        super().dispose()


class FileSessionProvider(ISessionProvider):
    def __init__(self, sessions_root: str, expired: int = None):
        self.sessions_root = os.path.abspath(os.path.join(sessions_root, 'restfx_sessions'))
        if not os.path.exists(self.sessions_root):
            os.makedirs(self.sessions_root)
            # print('mkdir:' + self.sessions_root)
        super().__init__(expired)

    def _get_session_path(self, session_id: str) -> str:
        # session_id 中可能存在 / 符号
        return os.path.join(self.sessions_root, session_id.replace('/', '_'))

    def _load_session(self, session_id: str):
        # print('Load session: ' + session_id)
        session_file = self._get_session_path(session_id)
        if not os.path.isfile(session_file):
            # print('The session currently loading not exists:' + session_file)
            return None

        if os.path.getsize(session_file) == 0:
            # print('The session currently loading is empty:' + session_file)
            return None

        import pickle

        with open(session_file, mode='rb') as fp:
            # noinspection PyBroadException
            try:
                session = pickle.load(fp)
                fp.close()
            except Exception:  # as e:
                # print('Failed to load session file %s: %s' % (session_file, repr(e)))
                fp.close()
                # 无法解析 session 文件
                # 说明session文件已经损坏，将其删除
                os.remove(session_file)
                return None

            setattr(session, '_update_watcher', self.set)
            setattr(session, '_drop_watcher', self.remove)
            return session

    def remove(self, session_id: str):
        # print('Remove session:' + session_id)
        session_file = self._get_session_path(session_id)
        if not self.exists(session_file):
            # print("The session to remove is not exists:" + session_file)
            return
        os.remove(session_file)

    def get_expired_session(self, time_before: float) -> List[str]:
        entities = os.listdir(self.sessions_root)

        sessions = []

        for entity in entities:
            last_access_time = os.path.getatime(self._get_session_path(entity))
            if time_before > last_access_time:
                sessions.append(entity)

        return sessions

    def get(self, session_id: str) -> Optional[HttpSession]:
        return self._load_session(session_id)

    def set(self, session: HttpSession):
        import pickle

        # print('Set session:' + session.id)
        with open(self._get_session_path(session.id), mode='wb') as fp:
            pickle.dump(session, fp, pickle.HIGHEST_PROTOCOL)
            fp.close()

    def exists(self, session_id: str):
        return os.path.isfile(self._get_session_path(session_id))

    def dispose(self):
        os.removedirs(self.sessions_root)
        super().dispose()


class MySQLSessionProvider(IDbSessionProvider):
    def __init__(self, pool_options: dict, table_name="restfx_sessions", expired: int = None):
        self.table_name = table_name

        super().__init__(pool_options, expired)

    def execute(self, sql: str, *args):
        import pymysql
        conn = self.connect()

        # print('[MySQLSessionProvider] ' + (sql % args))

        cursor = None
        try:
            cursor = conn.cursor(pymysql.cursors.DictCursor)
            rows = cursor.execute(sql, args)
            if sql.startswith('SELECT'):
                data = cursor.fetchall()
            else:
                data = None
        except Exception as e:
            print(str(e))
            return 0, None
        else:
            conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            conn.close()

        return rows, data

    def table_exists(self) -> bool:
        rows, _ = self.execute(
            """SELECT * FROM `information_schema`.`TABLES`
            WHERE `TABLE_SCHEMA`='{db_name}' AND `TABLE_NAME` ='{table_name}' LIMIT 1""".format(
                db_name=self.pool_option['database'],
                table_name=self.table_name
            ))
        return rows > 0

    def create_table(self):
        self.execute("""CREATE TABLE `{table_name}` (
        `id` VARCHAR(256) PRIMARY KEY NOT NULL,
        `creation_time` LONG NOT NULL,
        `last_access_time` LONG NOT NULL,
        `store` BLOB,
        INDEX `last_access`(`last_access_time`(8) ASC)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8""".format(table_name=self.table_name))

    def get_expired_session(self, time_before: float) -> List[str]:
        rows, data = self.execute(
            'SELECT `id` FROM `{table_name}` WHERE `last_access_time` < %s'.format(table_name=self.table_name),
            int(time_before))
        if rows == 0:
            return []
        return [item['id'] for item in data]

    def get(self, session_id: str) -> Optional[HttpSession]:
        rows, data = self.execute(
            "SELECT * FROM `{table_name}` WHERE `id`=%s LIMIT 1".format(table_name=self.table_name),
            session_id)
        if rows == 0:
            return None
        return self.parse(data[0])

    def upsert(self, session_id: str, creation_time: float, last_access_time: float, store: bytes):
        # 检查存储数据长度，如果大于 65535 则抛出异常
        store_len = len(store)
        max_value = 65535
        if store_len > max_value:
            raise RuntimeError('Entity is too large (%s) for session storage, the max value is %s.' % (
                store_len, max_value))

        self.execute("""INSERT INTO `{table_name}` VALUES(%s, %s, %s, %s) ON DUPLICATE KEY UPDATE
        last_access_time=%s, store=%s""".format(table_name=self.table_name),
                     session_id,
                     int(creation_time),
                     int(last_access_time),
                     store,
                     int(last_access_time),
                     store)

    def exists(self, session_id: str) -> bool:
        rows, _ = self.execute("""SELECT 1 FROM `{table_name}` WHERE `id`=%s limit 1""".format(
            table_name=self.table_name), session_id)
        return rows > 0

    def remove(self, session_id: str):
        self.execute("""DELETE FROM `{table_name}` WHERE `id`=%s""".format(table_name=self.table_name), session_id)

    def dispose(self):
        self.execute('TRUNCATE TABLE `{table_name}`'.format(table_name=self.table_name))
        super().dispose()



import base64


def enc_bytes(s: (str, bytes)) -> bytes:
    if isinstance(s, str):
        s = s.encode(encoding='utf8')
    return base64.b64encode(s)


def enc_str(s: (str, bytes)) -> str:
    return enc_bytes(s).decode(encoding='utf8')


def dec_bytes(s: (str, bytes)) -> bytes:
    if isinstance(s, str):
        s = s.encode(encoding='utf8')
    return base64.b64decode(s)


def dec_str(s: (str, bytes)) -> str:
    return dec_bytes(s).decode(encoding='utf8')



import os
from typing import Dict, Optional

from ..config import AppConfig
from ..http.response import NotFound, HttpResponse
from ..util import utils
from ..util.func_util import FunctionDescription


class RouteResolver:
    def __init__(self, config: AppConfig,
                 entry_cache: Dict[str, Optional[FunctionDescription]],
                 method: str,
                 entry: str):
        self.config = config
        self.entry_cache = entry_cache
        self.entry = entry
        self.method = method.lower()
        from ..util import Logger
        self.logger = Logger.get(config.app_id)

    def _get_module_abs_path(self, module_name):
        abs_path = os.path.join(self.config.ROOT, module_name)

        if os.path.isdir(abs_path):
            # 如果 module_name 是目录，那么就查找 __init__.py 是否存在
            self.logger.info('Entry "%s" is a package, auto load module "__init__.py"' % module_name)
            return '%s%s%s' % (module_name, os.path.sep, '__init__')

        if os.path.exists('%s.py' % abs_path):
            return module_name

        return None

    def resolve(self) -> [FunctionDescription, HttpResponse]:
        # entry 可能包含扩展名
        temp = self.entry.split('.')
        entry = temp[0]
        if len(temp) > 1:
            extname = temp[1]
        else:
            extname = None
        # 处理映射
        # 对应的模块(包或模块路径，使用 . 分隔）
        module_name = self.get_route_map(entry)

        if module_name is None:
            self.logger.warning(
                'Cannot find route "%s" in routes_map' % self.entry)
            return NotFound()

        func_name = self.method
        module_path = module_name.replace('.', os.path.sep)
        # 先检查完整路径是否存在
        module_abs_path = self._get_module_abs_path(module_path)
        if module_abs_path is None:
            # 再检查 append 函数是否存在
            # 将请求路径拆分成 entry 和 name
            temp = module_path.split(os.path.sep)
            name = temp.pop()
            module_path = os.path.sep.join(temp)
            module_abs_path = self._get_module_abs_path(module_path)

            if module_abs_path is None:
                return NotFound()

            module_name = module_abs_path.replace(os.path.sep, '.')

            # 如果指定了名称，那么就加上
            # 如：name = 'detail'
            #   func_name = get_detail
            func_name = '%s_%s' % (self.method, name.lower())

        # 完全限定名称
        fullname = '%s.%s' % (module_name, func_name)
        try:
            desc = self._get_handler_info(module_name, func_name, fullname)
        except Exception as e:
            message = 'Failed to load entry "%s": %s' % (self.entry, fullname)
            self.logger.error(message, e)
            return NotFound()

        # 检查 extname 是否一致
        if isinstance(desc, FunctionDescription) and desc.decorator['extname'] != extname:
            message = 'Failed to load entry "%s": extname "%s" is not exactly match with "%s"' % (
                self.entry, extname, desc.decorator['extname'])
            self.logger.warning(message)
            return NotFound()

        return desc

    def get_route_map(self, route_path):
        # 命中
        hit_route = None
        for root_path in self.config.routes_map:
            if route_path.startswith(root_path):
                hit_route = root_path, self.config.routes_map[root_path]
                break

        if hit_route is None:
            return None

        # 将请求路径替换为指定的映射路径
        return ('%s%s' % (hit_route[1], route_path[len(hit_route[0]):])).replace('/', '.').strip('.')

    def _get_handler_info(self, module_name, func_name, fullname) -> [FunctionDescription, HttpResponse]:
        # 缓存中有这个函数
        if fullname in self.entry_cache.keys():
            return self.entry_cache[fullname]

        # 缓存中没有这个函数，去模块中查找
        # ---------------

        try:
            entry_define = utils.load_module(module_name.replace('/', '.'))
        except Exception as e:
            message = 'Failed to load module "%s"' % module_name
            self.logger.error(message, e)
            return NotFound()

        # 模块中也没有这个函数
        if not hasattr(entry_define, func_name):
            # 函数不存在，更新缓存
            self.entry_cache[func_name] = None
            return NotFound()

        # 模块中有这个函数
        # 通过反射从模块加载函数
        func = getattr(entry_define, func_name)
        import inspect
        from . import Collector
        filename = inspect.getmodule(func).__file__
        decorator = Collector.get(self.config.app_id).resolve_routes(filename, func_name)

        if decorator is None:
            msg = '%s\n\tDecorator "@route" not found on function "%s", did you forgot it ?' % (
                utils.get_func_info(func),
                fullname
            )
            self.logger.warning(msg)
            # 没有配置装饰器@route，则认为函数不可访问，更新缓存
            self.entry_cache[func_name] = None
            return NotFound()

        func_desc = FunctionDescription(func)
        func_desc.decorator = decorator
        self.entry_cache[fullname] = func_desc

        return self.entry_cache[fullname]



import time


class HttpSession:
    def __init__(self, session_id, update_watcher, drop_watcher):
        """

        """
        self.id = session_id
        # 用于存放 session 数据项
        self.store = {}

        now = time.time()

        # 创建 session 的时间
        self.creation_time = now
        # 当前 session 对应的请求上下文最后被客户端访问的时间
        self.last_access_time = now
        # session 项变更的观察者
        self._update_watcher = update_watcher
        # session 销毁的观察者
        self._drop_watcher = drop_watcher
        # 当前是否已经被销毁
        self._destroyed = False
        # session 是否有变更
        # 默认值指定为 True，以在新建session后，能够被存储
        self._changed = True

    def __str__(self):
        return self.id

    def __getstate__(self):
        return {
            'id': self.id,
            'store': self.store,
            'creation_time': self.creation_time,
            'last_access_time': self.last_access_time,
        }

    def __setstate__(self, state):
        self.id = state['id']
        self.store = state['store']
        self.creation_time = state['creation_time']
        self.last_access_time = state['last_access_time']
        self._destroyed = False
        self._changed = False

    def get(self, key, default=None):
        """
        获取 key 所指定的项，当不存在时，返回 default 指定的值
        :param key:
        :param default:
        :return:
        """
        if self._destroyed:
            return None
        return self.store[key] if self.has(key) else default

    def set(self, key, value):
        """
        设置 key 指定的项值为 value
        :param key:
        :param value:
        :return:
        """
        if self._destroyed:
            return
        self.store[key] = value
        self._changed = True

    def has(self, key) -> bool:
        """
        指定的 key 是否存在于 session 中
        :param key:
        :return:
        """
        if self._destroyed:
            return False
        return key in self.store

    def remove(self, key):
        """
        从 session 中移除指定的 key
        :param key:
        :return:
        """
        if self._destroyed:
            return
        if self.has(key):
            del self.store[key]
            self._changed = True

    def flush(self):
        """
        立即将session中的变更存储，否则会在向浏览器发送响应时才会存储
        :return:
        """
        if not self._changed:
            return
        self._changed = False
        self._update_watcher(self)

    def clear(self):
        """
        清空所有项并销毁此 session
        :return:
        """
        if self._destroyed:
            return
        self._destroyed = True
        self.store.clear()
        self._drop_watcher(self.id)



import inspect
import json
import re
from collections import OrderedDict
from types import FunctionType

from restfx.session import HttpSession

from .utils import get_func_info
from ..http import HttpRequest


class ArgumentSpecification:
    """
    函数参数声明
    """

    def __init__(self, name: str, index: int):
        """

        :param name: 参数名称
        :param index: 参数在参数位置中的位置
        """
        self.name = name
        self.index = index
        # 是否是注入参数
        # 注入参数名称以 _ 开头
        self.is_injection = name[0] == '_'
        # 是否是可变参数
        self.is_variable = False
        # 是否有类型声明
        self.has_annotation = False
        # 是否有默认值
        self.has_default = False
        # 类型声明
        self.annotation = None
        # 类型是否是 tuple
        # 因为在处理类型时，会将 tuple 当作 list 对待
        self.is_tuple = False
        # 默认值
        self.default = None
        # 注释
        self.comment = None
        # 别名，当路由函数中声明的是 abc_def_xyz 时，自动处理为 abcDefXyz
        self.alias = None
        # TODO 校验器
        self.validator = None

        if self.is_injection:
            return

        if '_' not in self.name:
            return

        # 同时会移除所有的 _ 符号
        self.alias = re.sub('_+(?P<ch>.?)', ArgumentSpecification._get_parameter_alias, name)

    @property
    def annotation_name(self):
        return self.annotation.__name__ if self.has_annotation else None

    def __str__(self):
        name = self.name

        name = name if self.alias is None else '%s/%s' % (name, self.alias)

        arg_type = self.annotation.__name__ if self.has_annotation else None

        if self.has_default:
            default_value = "'%s'" % self.default if isinstance(self.default, str) else self.default
            return '%s: %s=%s' % (name, arg_type, default_value)

        return '%s: %s' % (name, arg_type)

    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if not isinstance(o, ArgumentSpecification):
                return json.JSONEncoder.default(self, o)
            return {
                'name': o.name,
                'index': o.index,
                'is_injection': o.is_injection,
                'is_variable': o.is_variable,
                'has_annotation': o.has_annotation,
                'has_default': o.has_default,
                'default': o.default,
                'annotation_name': o.annotation_name,
                'comment': o.comment,
                'alias': o.alias
            }

    @staticmethod
    def _get_parameter_alias(match):
        """

        :param match:
        :type match: Match
        :return:
        """
        ch = match.group('ch')
        return '' if ch is None else ch.upper()


class FunctionDescription:
    node_iter = None
    """
    :type: NodeVisitor
    """

    def __init__(self, func: FunctionType):
        self.func = func
        self.description = ''
        self.return_description = ''
        self.return_type = ''
        self.arguments = self._get_args()
        self.decorator = None

    def _get_args(self):
        """
        获取函数的参数列表（带参数类型）
        :return:
        """
        signature = inspect.signature(self.func)
        parameters = signature.parameters
        _empty = signature.empty

        documentation = self._get_func_docs()

        self.description = documentation['__func__']
        self.return_type = documentation['__rtype__']
        self.return_description = documentation['__return__']

        args = OrderedDict()
        index = 0
        for p in parameters.keys():
            parameter = parameters.get(p)
            spec = ArgumentSpecification(p, index)

            spec.is_variable = parameter.kind == parameter.VAR_KEYWORD

            spec.comment = documentation.get(p)
            index += 1
            # 类型
            annotation = parameter.annotation

            # 无效的类型声明 (意外的代码错误)
            if not inspect.isclass(annotation):
                source_lines = inspect.getsourcelines(self.func)
                line = source_lines[1]
                row = None
                for row in source_lines[0]:
                    if parameter.name in row:
                        break
                    line += 1
                msg = get_func_info(self.func)
                msg += 'Invalid type declaration for argument "%s":\n\t\t%s'
                # abort the execution
                raise Exception(msg % (parameter.name, row.strip()))

            default = parameter.default

            if default != _empty:
                spec.default = default
                spec.has_default = True

            # 有默认值时，若未指定类型，则使用默认值的类型
            if annotation == _empty:
                if default is not None and default != _empty:
                    spec.annotation = type(default)
                    spec.has_annotation = True
                elif p == 'request':
                    # 以下情况将设置为 HttpRequest 对象
                    # 1. 当参数名称是 request 并且未指定类型
                    # 2. 当参数类型是 HttpRequest 时 (不论参数名称，包括 request)
                    # 但是，参数名称是 request 但其类型不是 HttpRequest ，就会被当作一般参数处理
                    spec.annotation = HttpRequest
                    spec.has_annotation = True
                elif p == 'session':
                    # 以下情况将设置为 HttpSession 对象
                    # 1. 当参数名称是 session 并且未指定类型
                    # 2. 当参数类型是 HttpSession 时 (不论参数名称，包括 session)
                    # 但是，参数名称是 session 但其类型不是 HttpSession ，就会被当作一般参数处理
                    spec.annotation = HttpSession
                    spec.has_annotation = True
            else:
                spec.annotation = annotation
                spec.has_annotation = True

            # 如果声明的是元组，那么当成 list 来传
            if spec.has_annotation and spec.annotation == tuple:
                spec.annotation = list
                spec.is_tuple = True
            args[p] = spec

        return args

    def _get_func_docs(self):
        docs = {
            '__func__': '',
            '__return__': '',
            '__rtype__': None
        }

        # :type str
        doc_str = inspect.getdoc(self.func)
        if doc_str is None:
            return docs

        lines = doc_str.splitlines(False)

        buffer = []

        colon_found = False

        last_name = None

        for line in lines:
            if not line.startswith(':'):
                buffer.append(line)
                continue

            if not colon_found:
                docs['__func__'] = self.filter_doc_buffer(buffer)
                buffer.clear()
                colon_found = True

            if last_name:
                docs[last_name] = self.filter_doc_buffer(buffer)
                buffer.clear()
                last_name = None

            match = re.match(r':param\s+(?P<name>[\S]+):\s*(?P<comment>.*)$', line)

            if match:
                last_name = match.group('name')
                buffer.append(match.group('comment'))
                continue

            match = re.match(r':return:\s*(?P<comment>.*)$', line)

            if match:
                last_name = '__return__'
                buffer.append(match.group('comment'))
                continue

            match = re.match(r':rtype:\s*(?P<comment>.*)$', line)

            if match:
                last_name = '__rtype__'
                buffer.append(match.group('comment'))
                continue

        if last_name:
            docs[last_name] = self.filter_doc_buffer(buffer)
            buffer.clear()
        return docs

    @staticmethod
    def filter_doc_buffer(buffer):
        """
        移除注释中前置和后置的空行
        :param buffer:
        :return:
        """
        length = len(buffer)

        start = 0
        end = length

        data_range = range(length)

        for i in data_range:
            line = buffer[i].lstrip()
            if line:
                break
            start += 1

        if start == end:
            return ''

        for i in data_range:
            line = buffer[-i].lstrip()
            if line:
                break
            end -= 1

        temp = []
        # 提供代码段解析支持
        # 按缩进量处理
        found_code = False
        buffer2 = []
        is_first_line = True
        last_line_is_blank = False
        first_indent = 0
        for line in buffer[start:end]:
            stripped_line = line.strip()
            indent = len(line) - len(line.lstrip())
            # 当前行有缩进 (>=2)，就表示其为代码段
            if indent >= 2 or (found_code and not stripped_line):
                if not stripped_line:
                    if last_line_is_blank:
                        found_code = False
                        continue
                    last_line_is_blank = True
                else:
                    last_line_is_blank = False
                is_first_line = False
                if not first_indent:
                    first_indent = indent
                if not found_code:
                    # 重置临时数据
                    buffer2 = ['\n']
                    temp.append({
                        'type': 'code',
                        'lines': buffer2
                    })
                buffer2.append(line[first_indent:] + '\n')
                found_code = True
                continue

            first_indent = 0
            if is_first_line:
                temp.append({
                    'type': 'text',
                    'lines': buffer2
                })
                is_first_line = False
            elif found_code:
                # 重置临时数据
                buffer2 = []
                temp.append({
                    'type': 'text',
                    'lines': buffer2
                })

            if stripped_line:
                if last_line_is_blank is True:
                    # 这个判断用于合并连续的多个空行为一个
                    buffer2.append('\n')
                    last_line_is_blank = False
                buffer2.append(line)
            elif not found_code:
                # 忽略代码块后的第一个空行
                last_line_is_blank = True

            found_code = False
        # return ''.join(temp)
        return temp

    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, ArgumentSpecification):
                # noinspection PyCallByClass
                return ArgumentSpecification.JSONEncoder.default(self, o)
            if not isinstance(o, FunctionDescription):
                return json.JSONEncoder.default(self, o)
            return {
                'name': o.func.__name__,
                'description': o.description,
                'return_description': o.return_description,
                'return_type': o.return_type,
                'arguments': [o.arguments[arg] for arg in o.arguments]
            }



import json
from collections import OrderedDict
from enum import Enum
from functools import partial, wraps
from typing import Tuple, Union

from werkzeug.datastructures import MultiDict

from .parameter_interface import IParam
from .validator import Validator
from ..config import AppConfig
from ..http import BadRequest, HttpRequest, HttpResponse, JsonResponse, ServerError
from ..routes.meta import RouteMeta
from ..session import HttpSession
from ..util import Logger
from ..util.func_util import ArgumentSpecification, get_func_info


def route(module=None, name=None, extname=None, validators: Union[Tuple[Validator, ...], Validator] = None, **kwargs):
    """
    用于控制路由的访问权限，路由均需要添加此装饰器，若未添加，则不可访问
    用法：
    @route('用户管理', '编辑用户')
    def get(req):
        pass
    :param module: str 路由所属模块，一般在查询权限时会用到
    :param name: str 路由名称，一般在查询权限时会用到
    :param extname: 给路径指定一个扩展名，不能包含 . 符号
    :param validators: 指定参数应用的校验规则，每个需要校验的参数为元组的一个项
    """

    # 支持错误的元组写法: (aaa)
    # 正确写法应该为: (aaa,)
    if not isinstance(validators, tuple) and validators is not None:
        validators = (validators,)

    def invoke_route(handler):
        @wraps(handler)
        def caller(*args_def):
            # 参数长度不为 2 时，认为是用户调用
            if len(args_def) != 2:
                return handler(*args_def)

            # Http 请求对象
            # :type HttpRequest
            request = args_def[0]
            # 函数声明时定义的参数列表
            # :type OrderedDict
            handler_args = args_def[1]

            # 如果传入的参数第一个不是 request，第二个不是 OrderedDict，
            # 那么就认为是用户调用，而不是路由调用
            # 此时直接将原参数传给 handler 进行调用
            if not isinstance(request, HttpRequest) or not isinstance(handler_args, OrderedDict):
                return handler(*args_def)

            config = AppConfig.get(request.app_id)

            meta = RouteMeta(
                handler,
                handler_args,
                route_id='%s_%s' % (handler.__module__.replace('_', '__').replace('.', '_'), handler.__name__),
                module=module,
                name=name,
                extname=extname,
                method=request.method,
                path=request.path,
                kwargs=kwargs,
            )

            return _invoke_with_route(request, meta, config, validators)

        return caller

    return invoke_route


def validate_args(validators: Tuple[Validator, ...], args: dict):
    if not validators:
        return None

    if not args:
        return None

    param_name = None
    result = None
    for validator in validators:
        # noinspection PyProtectedMember
        param_name = validator._param_name
        result = validator.validate(args)
        if result is None:
            continue
        break
    if result is None:
        return result

    return 'Incorrect value of parameter "%s": %s' % (
        param_name,
        result
    )


def _invoke_with_route(request: HttpRequest, meta: RouteMeta, config: AppConfig, validators: tuple):
    handler_args = meta.handler_args
    func = meta.handler

    mgr = config.middleware_manager

    # 处理请求中的json参数
    _process_json_args(request, config)

    # 调用中间件，以处理请求
    result = mgr.handle_request(request, meta)

    # 使用函数代理，减少相同调用的参数传递
    handle_response = partial(mgr.handle_response, request, meta)
    wrap_response = partial(_wrap_http_response, mgr, request, meta)

    # 返回了 HttpResponse，直接返回此对象
    if isinstance(result, HttpResponse):
        return handle_response(result)

    # 返回了非 None，表示停止请求，并将结果作为路由的返回值
    if result is not None:
        return handle_response(wrap_response(result))

    # 有参数，自动从 queryString, POST 或 json 中获取
    # 匹配参数
    actual_args = _get_actual_args(request, func, handler_args, config)

    # 只有解析参数出错时才会返回 HttpResponse
    # 此时中止执行
    if isinstance(actual_args, HttpResponse):
        return handle_response(actual_args)

    # 调用中间件
    result = mgr.before_invoke(request, meta, actual_args)

    # 返回了 HttpResponse ， 直接返回此对象
    if isinstance(result, HttpResponse):
        return handle_response(result)

    # 返回了非 None，表示停止请求，并将结果作为路由的返回值
    if result is not None:
        return handle_response(wrap_response(result))

    # 执行校验
    result = validate_args(validators, actual_args)
    if result is not None:
        return BadRequest(result, content_type='text/plain')

    # 调用路由函数
    arg_len = len(handler_args)

    try:
        if arg_len == 0:
            result = func()
        else:
            result = func(**actual_args)
    except Exception as e:
        message = get_func_info(func)
        Logger.get(config.app_id).error(message, e)
        if config.debug:
            raise e
        from restfx.util import utils
        return handle_response(ServerError(message))

    return handle_response(wrap_response(result))


def _process_json_args(request: HttpRequest, config: AppConfig):
    """
    参数处理
    :return:
    """
    if request.content_type is None or 'application/json' not in request.content_type:
        return

    # 如果请求是json类型，就先处理一下

    body = request.data

    if not body:
        return

    try:
        request.BODY = json.loads(body.decode())
    except Exception as e:
        Logger.get(config.app_id).warning('Failed to deserialize request body: %s' % str(e))


def _get_parameter_str(args_def: OrderedDict):
    return ', '.join(filter(lambda n: n[0] != '_', [str(args_def[arg]) for arg in args_def]))


def _get_value(data: MultiDict, name: str, arg_spec: ArgumentSpecification):
    """

    :param data:
    :param name:
    :param arg_spec:
    :return: True 表示使用默认值 False 表示未使用默认值 None 表示无值
    """
    alias = arg_spec.alias
    # 提供对 a[]=1&a[]=2 的支持
    name_arr = name + '[]'
    alias_arr = alias + '[]' if alias else None
    is_arr = False
    if name in data:
        result = False, data.getlist(name)
    elif alias and alias in data:
        result = False, data.getlist(alias)
    elif name_arr in data:
        is_arr = True
        result = False, data.getlist(name_arr)
    elif alias_arr and alias_arr in data:
        is_arr = True
        result = False, data.getlist(alias_arr)
    elif arg_spec.has_default:
        # 使用默认值
        result = True, arg_spec.default
    else:
        # 缺少无默认值的参数
        result = None, None

    use_default, value = result

    if use_default is not False:
        return result

    if is_arr:
        # 请求参数本就为数组: param[]
        # 返回为 tuple 类型
        return use_default, tuple(value)

    # 当请求的不是数组时，使用最后一个值
    return use_default, value[-1]


def _get_enum_value(arg_spec, arg_value):
    # 此处的判断，始终忽略大小写
    tv = str(arg_value).lower()
    for enum_item in arg_spec.annotation:
        # 先判断值
        # 再判断名称
        if str(enum_item.value).lower() == tv or str(enum_item.name).lower() == tv:
            # 命中了此枚举项
            return enum_item
    # 没有命中
    return None


def _get_input_value(arg_spec, arg_name, arg_value, config):
    # 未指定类型
    if not arg_spec.has_annotation:
        return arg_value

    # 当值为 None 时，不作数据类型校验
    if arg_value is None:
        return arg_value

    # 检查类型是否一致 #

    # 类型一致，直接使用
    if isinstance(arg_value, arg_spec.annotation):
        return arg_value

    # 类型不一致，尝试转换类型
    if arg_spec.annotation is tuple and isinstance(arg_value, list):
        return tuple(arg_value)

    if arg_spec.annotation is list and isinstance(arg_value, tuple):
        return list(arg_value)

    # 当声明的参数类型是布尔类型时，收到的值可能是一个字符串（其值为 true 、 false）
    if arg_spec.annotation is bool and isinstance(arg_value, str):
        if arg_value == 'true':
            return True
        if arg_value == 'false':
            return False

        msg = 'Cannot parse value "%s" as type "%s for parameter %s". (expected: true/false)' % (
            (arg_value, arg_spec.annotation_name, arg_name))
        Logger.get(config.app_id).warning(msg)
        return BadRequest(msg)

    # 当声明的参数类型是枚举类型时，遍历枚举项，同时将枚举名称与值进行处理
    if issubclass(arg_spec.annotation, Enum):
        result = _get_enum_value(arg_spec, arg_value)
        if result is None:
            msg = 'Cannot parse value "%s" as type "%s" for parameter "%s".' % (
                arg_value, arg_spec.annotation_name, arg_name
            )
            Logger.get(config.app_id).warning(msg)
            return BadRequest(msg)

        return result

    # 声明的类型为 IParam，调用  parse 进行转换
    if issubclass(arg_spec.annotation, IParam):
        result = arg_spec.annotation.parse(arg_value)
        return result

    # 转换失败时，会抛出异常
    # noinspection PyBroadException
    try:
        # 当 arg_value 是字符串，arg_spec的类型是对象时，尝试解析成 json
        if arg_spec.annotation in (dict, list) and isinstance(arg_value, str):
            # noinspection PyBroadException
            try:
                arg_value = json.loads(arg_value)
            except Exception:
                # 此处的异常直接忽略即可
                msg = 'Cannot parse value "%s" as type "%s" for parameter "%s".' % (
                    arg_value, arg_spec.annotation_name, arg_name
                )
                Logger.get(config.app_id).warning(msg)
                return BadRequest(msg)

        # 类型一致，直接使用
        if isinstance(arg_value, arg_spec.annotation):
            # 如果原始声明是 tuple 类型，那么把 list 转换成 tuple
            # 虽然在发起请求的时候并不能指定为 tuple，但还是想兼容一下
            return tuple(arg_value) if arg_spec.is_tuple else arg_value
        return arg_spec.annotation(arg_value)
    except Exception:
        msg = 'Cannot parse value "%s" as type "%s" for parameter "%s".' % (
            arg_value, arg_spec.annotation_name, arg_name
        )
        Logger.get(config.app_id).warning(msg)
        return BadRequest(msg)


def _fill_session(request, func, arg_spec, arg_name, args_def, config):
    # 以下情况将传入 HttpSession 对象
    # 1. 当参数名称是 session 并且未指定类型
    # 2. 当参数类型是 HttpSession 时 (不论参数名称，包括 session)
    # 但是，参数名称是 session 但其类型不是 HttpSession ，就会被当作一般参数处理
    is_session = False
    if arg_name == 'session' and not arg_spec.has_annotation:
        is_session = True
    elif arg_spec.annotation == HttpSession:
        is_session = True

    if is_session:
        if request.session is None:
            msg = '%s\n\tParameter "%s" of type "HttpSession" is not available, ' \
                  'the value will always be "None", ' \
                  'please make sure that you have registered the Session-Middleware correctly, ' \
                  'Parameters: (%s)' % (
                      get_func_info(func),
                      arg_name,
                      _get_parameter_str(args_def)
                  )
            Logger.get(config.app_id).warning(msg)
        # 在 session 未启用时，其值为 None
        return request.session
    return None


def _fill_injections(request, func, injection_args, actual_args, config):
    # 填充注入参数
    for arg_name in injection_args:
        # 注入名称不包含前缀 _
        # 所以要 [1:]
        injection_name = arg_name[1:]
        # noinspection PyProtectedMember
        if injection_name in request._injections:
            # noinspection PyProtectedMember
            actual_args[arg_name] = request._injections[injection_name]
            continue

        if injection_name in config.injections:
            actual_args[arg_name] = config.injections[injection_name]
            continue

        msg = 'Injection name "%s" not found.' % injection_name
        if config.debug:
            msg = '%s\n\t%s' % (get_func_info(func), msg)
        Logger.get(config.app_id).error(msg)
        return ServerError(msg) if config.debug else ServerError()
    return None


def _get_variable_args(arg_source, used_args):
    # 填充可变参数
    variable_args = {}
    for item in arg_source:
        if item in used_args:
            continue
        value = arg_source.getlist(item)
        # 当值长度大于1时，使用 tuple
        # 或者 item 以 [] 结尾
        if item.endswith('[]'):
            variable_args[item[:-2]] = tuple(value)
            continue
        variable_args[item] = tuple(value) if len(value) > 1 else value[-1]

    return variable_args


def _get_actual_args(request: HttpRequest, func, args_def: OrderedDict, config: AppConfig) -> dict or HttpResponse:
    actual_args = {}

    # 已使用的参数名称，用于后期填充可变参数时作排除用
    used_args = []
    # 是否声明了可变参数
    has_variable_args = False

    # 合并参数成一个对象
    # 注意：不同来源（get/post）的参数会被覆盖 (post 覆盖 get) 的值
    arg_source = MultiDict(request.GET)
    arg_source.update(request.POST)
    if isinstance(request.BODY, dict):
        arg_source.update(request.BODY)
    if isinstance(request.FILES, dict):
        arg_source.update(request.FILES)

    # 声明的注入参数集合
    injection_args = []

    for arg_name in args_def.keys():
        arg_spec = args_def.get(arg_name)

        if arg_spec.is_injection:
            injection_args.append(arg_name)
            used_args.append(arg_name)
            continue

        # 如果是可变参数：如: **kwargs
        # 设置标记，以在后面进行填充
        if arg_spec.is_variable:
            has_variable_args = True
            continue

        # 以下情况将传入 HttpRequest 对象
        # 1. 当参数名称是 request 并且未指定类型
        # 2. 当参数类型是 HttpRequest 时 (不论参数名称，包括 request)
        # 但是，参数名称是 request 但其类型不是 HttpRequest ，就会被当作一般参数处理
        if arg_name == 'request' and not arg_spec.has_annotation:
            actual_args[arg_name] = request
            continue
        if arg_spec.annotation == HttpRequest:
            actual_args[arg_name] = request
            continue

        # 填充 session
        sess = _fill_session(request, func, arg_spec, arg_name, args_def, config)
        if sess:
            actual_args[arg_name] = sess
            continue

        use_default, arg_value = _get_value(arg_source, arg_name, arg_spec)

        # 未找到参数
        if use_default is None:
            msg = 'Missing required argument "%s".' % arg_name
            Logger.get(config.app_id).warning(msg)
            return BadRequest(msg)

        # 使用默认值
        if use_default is True:
            actual_args[arg_name] = arg_value
            used_args.append(arg_name)
            continue

        val = _get_input_value(arg_spec, arg_name, arg_value, config)

        if isinstance(val, HttpResponse):
            return val
        actual_args[arg_name] = val
        used_args.append(arg_name)

    result = _fill_injections(request, func, injection_args, actual_args, config)
    if isinstance(result, HttpResponse):
        return result

    variable_args = _get_variable_args(arg_source, used_args)

    # 没有可变参数
    if not variable_args:
        return actual_args

    # 有可变参数，并且指定了 kwargs
    if has_variable_args:
        actual_args.update(variable_args)
        return actual_args

    # 有可变参数，并且未指定 kwargs
    # 未启用严格模式
    if config.strict_mode is not True:
        return actual_args

    # 启用了严格模式
    # 返回 400 响应
    variable_arg_keys = variable_args.keys()
    msg = 'Unknown argument(s) found: "%s", Parameters: (%s).' \
          % (','.join(variable_arg_keys), _get_parameter_str(args_def))
    Logger.get(config.app_id).warning(msg)
    if not config.debug:
        msg = 'Unknown argument(s) found: %s.' % ','.join(variable_arg_keys)
    return BadRequest(msg)


def _wrap_http_response(mgr, request, meta, data):
    """
    将数据包装成 HttpResponse 返回
    :param data:
    :return:
    """

    # 调用中间件，处理返回函数
    data = mgr.after_return(request, meta, data)

    if data is None:
        return HttpResponse()

    if isinstance(data, HttpResponse):
        return data

    if isinstance(data, bool):
        return HttpResponse('true' if bool else 'false', content_type='text/plain')

    if isinstance(data, (dict, list, set, tuple)):
        return JsonResponse(data)

    if isinstance(data, str):
        return HttpResponse(data.encode(), content_type='text/plain')

    if isinstance(data, bytes):
        return HttpResponse(data, content_type='application/octet-stream')

    return HttpResponse(str(data).encode(), content_type='text/plain')



import inspect
import re


def load_module(module_name: str, level=0):
    """
    加载模块
    :param level:
    :param module_name:
    :return:
    """
    # __import__ 自带缓存
    module = __import__(module_name, level=level, globals={
        **globals(),
        **locals()
    })
    temp = module_name.split('.')
    if len(temp) > 1:
        for seg in temp[1:]:
            module = getattr(module, seg)
    return module


def get_func_info(func):
    source_lines = inspect.getsourcelines(func)
    line = source_lines[1]
    return 'File "%s", line %d, code %s' % (
        inspect.getmodule(func).__file__,
        line,
        func.__name__
    )


def get_ip_list():
    """
    读取本机的IP地址列表
    :return:
    """
    import socket

    addresses = socket.getaddrinfo(socket.gethostname(), None, socket.AF_INET)

    ips = ['127.0.0.1']
    for address in addresses:
        ip = address[4][0]
        if ip not in ips:
            ips.append(ip)

    return ips


def get_exception_info(e: Exception, message: str = None):
    messages = []
    if e:
        tb = e.__traceback__

        while tb:
            msg = str(tb.tb_frame.f_code)
            match = re.match(r'^<code object (?P<name>(<module>|[a-zA-Z0-9_]+?)) at.+?, file (?P<file>(.+?)),.+$', msg)
            if match:
                msg = 'File %s, line %s, code %s' % (match.group('file'), tb.tb_frame.f_lineno, match.group('name'))
            messages.append(msg)

            tb = tb.tb_next

    if message:
        messages.append(message)
    else:
        messages.append(str(e))

    return '\n\t'.join(messages)



import re

import setuptools

with open("src/restfx/__meta__.py", encoding="utf8") as f:
    version = re.search(r'version = ([\'"])(.*?)\1', f.read()).group(2)

setuptools.setup(
    name='restfx',
    version=version,
    install_requires=[
        'Werkzeug>=1.0.0'
    ]
)



import re
from typing import Union

from restfx.http import HttpFile


class DefaultValidators:
    @classmethod
    def range(cls, min_, max_, include_min, include_max, msg, param_name: str, args: dict):
        if not min_ and not max_:
            return True

        value = args.get(param_name)

        if isinstance(value, (tuple, list, str)):
            temp = len(value)
        elif isinstance(value, HttpFile):
            temp = value.size
        elif value is None:
            # 值为 None 时，认为其长度为0（或者值为0）
            temp = 0
        else:
            temp = value

        if include_min and include_max:
            result = min_ <= temp <= max_
        elif include_min:
            result = min_ <= temp < max_
        else:
            result = min_ < temp <= max_

        return None if result else msg

    @classmethod
    def enums(cls, enums, msg, param_name: str, args: dict):
        value = args.get(param_name)
        return None if value in enums else msg

    @classmethod
    def match(cls, other_param_name, msg, param_name: str, args: dict):
        value = args.get(param_name)
        other_value = args.get(other_param_name)
        return None if value == other_value else msg

    @classmethod
    def regex(cls, pattern, flags, msg, param_name: str, args: dict):
        import re
        value = args.get(param_name)
        expr = re.compile(pattern, flags)
        return None if expr.fullmatch(value) else msg

    @classmethod
    def file(cls, file_ext, file_mime, msg, param_name: str, args: dict):
        value = args.get(param_name)
        """
        :type: HttpFile
        """
        if file_ext:
            if not value.filename.lower().endswith(file_ext.lower().split(';')):
                return msg
        if file_mime:
            if not value.mimetype.lower().endswith(file_mime.lower().split(';')):
                return msg
        return None

    @classmethod
    def empty(cls, allow, msg, param_name: str, args: dict):
        """
        返回 true 表示继续后续的校验

        返回 false 表中止后续校验

        返回 其它(字符串)表示校验失败
        :param allow:
        :param msg:
        :param param_name:
        :param args:
        :return:
        """
        value = args.get(param_name)

        if value:
            return True

        if allow:
            return False

        return msg


class Validator:
    """
    校验描述器，目前支持对 str, int, float, tuple, list, HttpFile 进行校验

    如果需要自定义校验器，那么请继承此类，添加自定义的方法即可

    在自定义的方法中，需要调用self._append_rule() 方法，以将自定义的校验加入到校验链中

    注意：自定义的方法必须返回 self，以支持链式调用
    """

    def __init__(self, param_name: str):
        self._param_name = param_name
        # 其内存储的项类型为 tuple 仅包含两项
        # 项的第一个值为 校验函数对象
        # 项的第二个值为 校验函数参数 (tuple)
        self._rule_chain = []
        """
        :type: List[Tuple(FunctionType, tuple)]
        """

    def _append_rule(self, val_fn, val_args: tuple):
        """
        向校验链中追加校验
        :param val_fn:
        :param val_args:
        :return:
        """
        self._rule_chain.append((val_fn, val_args))

    def range(self, min_: Union[int, float] = 0, max_: Union[int, float] = 0,
              include_min=True, include_max=True, msg='Value out of range'):
        """
        用于限制 字符串的长度 以及 int/float 的值范围

        若 min 和 max 都为 0, 那么表示不限制

        若都不为 0 ，表示限制输入数值范围或字符串长度

        当仅指定了 min 时，表示仅限制最小值/最小长度

        当仅指定了 最大值 时，表示仅限制最大值/最大长度

        在校验时，会包含端点值
        :param min_:
        :param max_:
        :param include_min: 是否包含最小值
        :param include_max: 是否包含最大值
        :param msg: 校验失败时的提示消息
        :return:
        """
        if min_ or max_:
            self._append_rule(DefaultValidators.range, (min_, max_, include_min, include_max, msg))
        return self

    def enum(self, enums: tuple, msg='Value out of range'):
        """
        用于指定允许可以使用的值枚举。但更建议将参数类型指定为 枚举类型
        :param enums:
        :param msg:
        :return:
        """
        if enums:
            self._append_rule(DefaultValidators.enums, (enums, msg))
        return self

    def when(self, **kwargs):
        """
        当满足指定的条件时执行校验
        :param kwargs: 指定字段和值，如果值为 tuple ，表示字段值在某范围内
        :return:
        """
        raise NotImplemented()

    def empty(self, allow=True, msg='Value must not be empty'):
        """
        是否允许输入为空 (包括空串，空列表，空元组)，这应该是校验的第一个接口
        :return:
        """
        self._append_rule(DefaultValidators.empty, (allow, msg))
        return self

    def regex(self, pattern: str = None, flags=0, msg='Invalid Value'):
        """
        仅应对字符串生效
        :param pattern:
        :param flags:
        :param msg:
        :return:
        """
        if pattern:
            self._append_rule(DefaultValidators.regex, (pattern, flags, msg))
        return self

    def ip(self, v4=True, v6=False, msg='Not a valid IP address'):
        """

        :param v4: 判断是否为 IPv4
        :param v6: 判断是否为 IPv6。保留参数，暂不支持
        :param msg:
        :return:
        """
        if v4:
            pattern = r'^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$'
        elif v6:
            pattern = None
        else:
            return self

        flags = 0
        self._append_rule(DefaultValidators.regex, (pattern, flags, msg))
        return self

    def email(self, msg='Not a valid Email'):
        pattern = r'^([a-z0-9_.-]+)@([\da-z.-]+)\.([a-z.]{2,6})$'
        flags = re.IGNORECASE
        self._append_rule(DefaultValidators.regex, (pattern, flags, msg))
        return self

    def match(self, other_param_name: str, msg='The two value not match'):
        """
        判断两个参数的值相同

        支持基本类型的比较：int float bool str list tuple dict
        :param other_param_name:
        :param msg:
        :return:
        """
        self._append_rule(DefaultValidators.match, (other_param_name, msg))
        return self

    def file(self, ext=None, mime=None, msg='Invalid file type'):
        """
        校验文件类型
        :param ext: 指定文件的扩展名，多个使用 ; 分隔
        :param mime: 指定文件的 mime，多个使用 ; 分隔
        :param msg:
        :return:
        """
        if ext or mime:
            self._append_rule(DefaultValidators.file, (ext, mime, msg))
        return self

    def validate(self, args: dict):
        """
        执行参数校验。
        :param args:
        :return: 返回 None 表示校验通过，返回字符串为校验失败的消息
        """
        common_args = [self._param_name, args]
        for fn, pre_args in self._rule_chain:
            result = fn(*pre_args, *common_args)
            if result is False and fn == DefaultValidators.empty:
                # 允许为空，并且值为空，不再进行后续的校验
                return None
            if isinstance(result, str):
                return result

        return None

    def __str__(self):
        return '<V %s=%s>' % (self._param_name, ','.join([
            '%s(%s)' % (rule.func.__name__, ','.join([str(i) for i in rule.args]))
            for rule in self._rule_chain
        ]))



from logging import getLogger

from restfx import __meta__
from restfx.util import utils

logger = getLogger(__meta__.name)


class Logger:
    _LOGGERS = {}
    colors = {
        'error': '\033[1;31m',
        'warning': '\033[1;33m',
        'debug': '\033[1;37m'
    }

    def __init__(self, app_id: str):
        self.app_id = app_id
        self._LOGGERS[app_id] = self
        self.custom_logger = None

    @classmethod
    def print(cls, level, message):

        if level not in cls.colors:
            print('%s' % message)
            return

        print('%s%s%s' % (cls.colors[level], message, '\033[0m'))

    def log(self, level, message, e=None):
        if self.custom_logger is not None:
            # noinspection PyUnresolvedReferences
            self.custom_logger(level, message, e)
            return
        getattr(logger, level)(message)
        # self.print(level, message)

    def debug(self, message):
        self.log('debug', message)

    def info(self, message):
        self.log('info', message)

    def warning(self, message):
        self.log('warning', message)

    def error(self, message, e=None):
        # self.log('error', utils.get_exception_info(e, message))
        self.log('error', message)

    @classmethod
    def remove(cls, app_id: str):
        if app_id in cls._LOGGERS:
            del cls._LOGGERS[app_id]

    @classmethod
    def get(cls, app_id: str):
        _logger = cls._LOGGERS.get(app_id, None)
        if _logger is None:
            _logger = Logger(app_id)
        return _logger



import time
import uuid

from ...config import AppConfig
from ...middleware.interface import MiddlewareBase
from ...session.interfaces import ISessionProvider
from ...util import md5, b64


class SessionMiddleware(MiddlewareBase):
    def __init__(self, provider: ISessionProvider,
                 secret=None,
                 maker=None,
                 cookie_name='sessionid',
                 cookie_max_age=None,
                 cookie_expires=None,
                 cookie_path="/",
                 cookie_domain=None,
                 cookie_secure=False,
                 cookie_samesite=None,
                 cookie_httponly=True
                 ):
        """

        :param provider:
        :param secret: 用于加密 session id 的密钥，设置为 None 时，将使用 app_id
        :param maker:session id 的创建算法:
            设置为 None 时表示使用默认的加密算法
            设置为函数表示自定义算法
        :param cookie_name:
        :param cookie_max_age:
        :param cookie_expires:
        :param cookie_path:
        :param cookie_domain:
        :param cookie_secure:
        :param cookie_samesite:
        :param cookie_httponly:
        """
        assert isinstance(provider, ISessionProvider)
        self.maker = maker
        self.secret = secret
        self.secret_bytes = None
        self.provider = provider
        self.cookie_name = cookie_name
        self.cookie_max_age = cookie_max_age
        self.cookie_expires = cookie_expires
        self.cookie_path = cookie_path
        self.cookie_domain = cookie_domain
        self.cookie_secure = cookie_secure
        self.cookie_samesite = cookie_samesite
        self.cookie_httponly = cookie_httponly

    @staticmethod
    def default_maker():
        return uuid.uuid4().hex

    def newid(self):
        sid_bytes = md5.hash_bytes(self.maker() if self.maker else self.default_maker())
        return b64.enc_str(sid_bytes)

    def decode(self, sid):
        # noinspection PyBroadException
        try:
            sid_bytes = b64.dec_bytes(sid)
            # 使用 secret 解密
            result = bytearray()
            for i in range(16):
                result.append(sid_bytes[i] ^ self.secret_bytes[i])

            return b64.enc_str(result)
        except Exception:
            # 解码失败，此 id 非法
            return None

    def on_startup(self, app):
        if self.secret is None:
            self.secret = app.id
        self.secret_bytes = md5.hash_bytes(self.secret)

    def process_request(self, request, meta):
        # 当指定了 session=False 时，表示此请求此路由时不需要创建 session
        if not meta.get('session', True):
            return
        config = AppConfig.get(request.app_id)
        if config is None or self.provider is None:
            return

        client_session_id = request.cookies.get(self.cookie_name)
        # 客户端无 session_id
        if not client_session_id:
            request.session = self.provider.create(self.newid())
            return

        # 解码 session_id
        session_id = self.decode(client_session_id)
        if not session_id:
            # session id 非法，新创建一个
            request.session = self.provider.create(self.newid())
            return

        # 尝试根据客户端的 session_id 获取 session
        request.session = self.provider.get(session_id)

        # session 已经过期或session被清除
        if request.session is None:
            request.session = self.provider.create(self.newid())
            return

        now = time.time()
        # session 过期
        if self.provider.is_expired(request.session):
            request.session = self.provider.create(self.newid())
            return

        # 修改已经存在 session 的最后访问时间
        request.session.last_access_time = now

    def on_leaving(self, request, response):
        # 在响应结束时才写入，以减少 IO
        if not request.session:
            return
        request.session.flush()
        # 使用 secret 加密
        sid_bytes = b64.dec_bytes(request.session.id)
        result = bytearray()
        for i in range(16):
            result.append(sid_bytes[i] ^ self.secret_bytes[i])
        # 加密后的 sid
        sid = b64.enc_str(result)
        response.set_cookie(self.cookie_name,
                            sid,
                            max_age=self.cookie_max_age,
                            expires=self.cookie_expires,
                            path=self.cookie_path,
                            domain=self.cookie_domain,
                            secure=self.cookie_secure,
                            httponly=self.cookie_httponly,
                            samesite=self.cookie_samesite,
                            )

    def __del__(self):
        del self.provider



import os
from collections import OrderedDict

from ..config import AppConfig
from ..http import HttpResponse, JsonResponse, BadRequest
from ..util.func_util import FunctionDescription


class ApiPage:
    def __init__(self, config: AppConfig):
        self.config = config
        # 开发模式的模块缓存，用于API列表
        self.routes_cache = {}
        # API列表页面缓存
        self.api_page_html_cache = ''

    def dispatch(self, request):
        if not self.config.api_page_enabled:
            from ..util import Logger
            Logger.get(self.config.app_id).info(
                'API page is disabled, '
                'use "App(..., api_page_enabled=True, ...)" to enable it.')
            return HttpResponse(status=404)

        if request.method == 'GET':
            return self.do_get()

        if request.method == 'POST':
            if 'export' in request.GET and request.GET['export'] == 'md':
                return self.do_export(request)
            return self.do_post()

        return HttpResponse(status=405)

    def do_get(self):
        if not self.api_page_html_cache or not self.config.api_page_cache:
            with open(os.path.join(os.path.dirname(__file__),
                                   '../internal_assets/templates/api_page.html'),
                      encoding='utf-8') as fp:
                lines = fp.readlines()
                self.api_page_html_cache = ''.join(lines)
                fp.close()
        return HttpResponse(self.api_page_html_cache, content_type='text/html')

    def do_post(self):
        if not self.routes_cache or not self.config.api_page_cache:
            from . import Collector
            routes = Collector.get(self.config.app_id).collect(self.config.routes_map)
            addition_func = self.config.api_page_addition

            for route in routes:
                # 附加信息
                if addition_func is not None:
                    route['addition_info'] = addition_func(route)

                # 移除参数上的其它信息 HttpRequest _xxx HttpSession
                raw_args = route['handler_info'].arguments
                new_args = OrderedDict()

                from ..http import HttpRequest
                from ..session import HttpSession

                for arg_name in raw_args:
                    arg = raw_args.get(arg_name)
                    if arg.is_injection:
                        continue
                    if arg.annotation == HttpRequest:
                        continue
                    if arg.annotation == HttpSession:
                        continue
                    new_args.setdefault(arg_name, arg)

                route['handler_info'].arguments = new_args

                # 移除 kwargs，以避免额外数据带来的数据传输消耗
                if 'kwargs' in route:
                    del route['kwargs']

            self.routes_cache = routes

        from restfx import __meta__
        return JsonResponse({
            'meta': {
                'name': __meta__.name,
                'version': __meta__.version,
                'url': __meta__.website
            },
            'name': self.config.api_page_name,
            'expanded': self.config.api_page_expanded,
            'routes': self.routes_cache,
        }, encoder=FunctionDescription.JSONEncoder)

    def do_export(self, request):
        if 'data' not in request.POST:
            return BadRequest()

        content = request.POST['data']

        # noinspection PyBroadException
        try:
            from restfx.util import b64
            content = b64.dec_bytes(content)
        except Exception:
            # This exception can be ignored safety
            return BadRequest()

        import time

        from restfx.http import FileResponse
        return FileResponse(content,
                            content_type='application/markdown;charset=utf-8',
                            request=request,
                            attachment='%s[%s].md' % (
                                self.config.api_page_name,
                                time.strftime('%Y%m%I%H%M%S'))
                            )
